from typing import Optional

from ewokscore import Task
from ewokscore.model import BaseOutputModel
from pydantic import Field

from ...io import hdf5
from .. import hdf5_utils
from .. import nexus_utils
from ..models import FluoInputs


class Inputs(FluoInputs):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )

    output_root_uri: str = Field(
        description="Output root URI.",
        examples=[
            "/results/dataset.h5",
            "/results/dataset.h5::/1.1",
            "/results/dataset.h5::/1.1/merge",
        ],
    )

    xrf_results_uri: Optional[str] = Field(
        default=None,
        description="Previous XRF NXprocess URI to merge the raw Bliss counters with.",
        examples=["/results/dataset.h5::/1.1/sum/results"],
    )


class Outputs(BaseOutputModel):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )
    output_root_uri: str = Field(
        description="Output NXentry URI.", examples=["/results/dataset.h5::/1.1"]
    )
    xrf_results_uri: str = Field(
        description="Output NXprocess results URI.",
        examples=["/results/dataset.h5::/1.1/merge/results"],
    )


class ExtractRawCounters(Task, input_model=Inputs, output_model=Outputs):
    """Extract raw single-scan counters and save them like XRF results."""

    def run(self):
        start_time = nexus_utils.now()
        bliss_scan_uri = self.inputs.bliss_scan_uri
        output_root_uri = self.inputs.output_root_uri
        previous_xrf_results_uri = self.inputs.xrf_results_uri

        input_file, scan_h5path = hdf5.split_h5uri(self.inputs.bliss_scan_uri)
        if previous_xrf_results_uri:
            default_nprocess_name = "merge"
        else:
            default_nprocess_name = "raw"

        with nexus_utils.save_in_ewoks_process(
            output_root_uri,
            start_time,
            process_config=dict(),
            default_levels=(scan_h5path, default_nprocess_name),
        ) as (process_group, already_existed):
            if already_existed:
                merged_xrf_results = process_group["results"]
            else:
                merged_xrf_results = process_group.create_group("results")
                merged_xrf_results.attrs["NX_class"] = "NXcollection"

                with hdf5.FileReadAccess(input_file) as h5file:
                    _link_raw_counters(
                        h5file[scan_h5path], merged_xrf_results, "rawcounters"
                    )

                    if previous_xrf_results_uri:
                        input_file, parent_path = hdf5.split_h5uri(
                            previous_xrf_results_uri
                        )
                        with hdf5.FileReadAccess(input_file) as h5file:
                            _link_xrf_results(h5file[parent_path], merged_xrf_results)

            self.outputs.xrf_results_uri = (
                f"{merged_xrf_results.file.filename}::{merged_xrf_results.name}"
            )

        self.outputs.bliss_scan_uri = bliss_scan_uri
        self.outputs.output_root_uri = output_root_uri


def _link_raw_counters(
    raw_scan: hdf5.GroupType, process_group: hdf5.GroupType, name: str
) -> None:
    destination = nexus_utils.create_nxdata(process_group, name)
    measurement = raw_scan["measurement"]
    signals = list()
    for name, dset in measurement.items():
        if dset.ndim == 1:
            hdf5_utils.create_hdf5_link(destination, name, dset)
            signals.append(name)
    nexus_utils.set_nxdata_signals(destination, signals=signals)


def _link_xrf_results(
    xrf_results: hdf5.GroupType, process_group: hdf5.GroupType
) -> None:
    for name, group in xrf_results.items():
        hdf5_utils.create_hdf5_link(process_group, name, group)



from dataclasses import dataclass
from typing import List
from typing import NamedTuple
from typing import Optional

from ...io.hdf5 import join_h5url
from ..pymca_config import PyMcaXrfConfiguration


class FitUris(NamedTuple):
    bliss_scan_uris: List[str]
    detector_name: str
    output_root_uri: str
    xrf_spectra_uri_template: str = "instrument/{}/data"
    energy_uri_template: str = "instrument/positioners/{}"
    energy_name: Optional[str] = None
    process_uri_template: str = "fit/{}"

    @property
    def xrf_spectra_uris(self) -> List[str]:
        template = self.xrf_spectra_uri_template
        return [
            f"{bliss_scan_uri}/{template.format(self.detector_name)}"
            for bliss_scan_uri in self.bliss_scan_uris
        ]

    @property
    def output_parent_uri(self) -> str:
        output_uri_template = join_h5url(
            self.output_root_uri, self.process_uri_template
        )
        return output_uri_template.format(self.detector_name)

    @property
    def beam_energy_uris(self) -> List[Optional[str]]:
        if not self.energy_name:
            return [None] * len(self.bliss_scan_uris)
        template = self.energy_uri_template
        return [
            f"{bliss_scan_uri}/{template.format(self.energy_name)}"
            for bliss_scan_uri in self.bliss_scan_uris
        ]


@dataclass
class ProgressStats:
    seconds_read: float = 0.0
    seconds_process: float = 0.0
    seconds_write: float = 0.0


class FitTask(NamedTuple):
    xrf_spectra_uri: str
    beam_energy_uri: Optional[str]
    start_in: int
    stop_in: int
    start_out: int
    stop_out: int
    output_parent_uri: str
    progress: ProgressStats

    @property
    def num_spectra(self) -> int:
        return self.stop_in - self.start_in


class FitArguments(NamedTuple):
    configuration: PyMcaXrfConfiguration
    quantification: Optional[bool]
    individual_weights: Optional[bool]
    positive_peak_areas: Optional[bool]
    energy_multiplier: Optional[bool]
    fast_fitting: Optional[bool]
    native_fitting: Optional[bool]
    native_legacy_fitting: Optional[bool]
    diagnostics: Optional[bool]



from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Sequence
from typing import Union

from ewokscore import Task
from ewokscore.model import BaseOutputModel
from pydantic import Field

from ...io import hdf5
from ...math.regular_grid import ScatterDataInterpolator
from ...math.rounding import round_to_significant
from .. import nexus_utils
from ..models import FluoInputs
from ..positioner_utils import get_primary_beam_energy_suburi
from ..xrf_results import get_xrf_result_groups
from . import positioners_utils


class Inputs(FluoInputs):
    bliss_scan_uris: List[str] = Field(
        description="Bliss scan URI's.",
        examples=[["/data/dataset.h5::/1.1", "/data/dataset.h5::/2.1"]],
    )

    output_root_uri: str = Field(
        description="Output root URI.",
        examples=[
            "/results/dataset.h5",
            "/results/dataset.h5::/1.1",
            "/results/dataset.h5::/1.1/regrid",
        ],
    )

    xrf_results_uri: Optional[str] = Field(
        description="Previous XRF NXprocess URI to regrid.",
        examples=["/results/dataset.h5::/1.1/norm/results"],
    )

    stack_positioner: Optional[str] = Field(default=None)
    positioners: Union[List[str], str, None] = Field(default=None)
    positioner_uri_template: str = Field(default="measurement/{}")
    ignore_positioners: Optional[List[str]] = Field(default=None)
    interpolate: Literal["nearest", "linear", "cubic"] = Field(default="nearest")
    resolution: Optional[dict] = Field(default=None)

    axes_units: Dict[str, str] = Field(
        default_factory=dict,
        description="Axes units to be used when missing.",
        examples=[{"nsy": "mm", "nsz": "mm", "nspy": "um", "nspz": "um"}],
    )


class Outputs(BaseOutputModel):
    bliss_scan_uris: List[str] = Field(
        description="Bliss scan URI's.",
        examples=[["/data/dataset.h5::/1.1", "/data/dataset.h5::/2.1"]],
    )

    output_root_uri: str = Field(
        description="Output NXentry URI.", examples=["/results/dataset.h5::/1.1"]
    )
    xrf_results_uri: str = Field(
        description="Output NXprocess results URI.",
        examples=["/results/dataset.h5::/1.1/regrid/results"],
    )


class RegridXrfResultsStack(Task, input_model=Inputs, output_model=Outputs):
    """Regrid XRF stack results on a regular grid by interpolation."""

    def run(self):
        start_time = nexus_utils.now()

        with nexus_utils.save_in_ewoks_subprocess(
            self.inputs.output_root_uri,
            start_time,
            {},
            default_levels=("results", "regrid"),
        ) as (regrid_results, already_existed):
            if not already_existed:
                self._regrid(regrid_results)

            self.outputs.xrf_results_uri = (
                f"{regrid_results.file.filename}::{regrid_results.name}"
            )
        self.outputs.bliss_scan_uris = self.inputs.bliss_scan_uris
        self.outputs.output_root_uri = self.inputs.output_root_uri

    def _regrid(self, regrid_results):
        bliss_scan_uris = self.inputs.bliss_scan_uris
        xrf_results_uri = self.inputs.xrf_results_uri
        ignore_positioners = self.inputs.ignore_positioners
        position_suburis = self._get_position_suburis(
            bliss_scan_uris, ignore_positioners=ignore_positioners
        )
        stack_suburi = self._get_stack_suburi(bliss_scan_uris)
        interpolate = self.inputs.interpolate
        resolution = self.inputs.resolution
        axes_units = self.inputs.axes_units

        xrf_results_filename, xrf_results_h5path = hdf5.split_h5uri(xrf_results_uri)

        with hdf5.FileReadAccess(xrf_results_filename) as xrf_results_file:
            xrf_results_grp = xrf_results_file[xrf_results_h5path]
            if not hdf5.is_group(xrf_results_grp):
                raise TypeError(f"'{xrf_results_h5path}' must be a group")

            nxdata_groups = get_xrf_result_groups(xrf_results_grp)

            coordinates, grid_names, grid_units = (
                positioners_utils.read_position_suburis(
                    bliss_scan_uris[0], position_suburis, axes_units=axes_units
                )
            )
            if resolution:
                resolution = [resolution[name] for name in grid_names]
            else:
                resolution = None
            interpolator = ScatterDataInterpolator(
                coordinates,
                grid_names,
                grid_units,
                method=interpolate,
                resolution=resolution,
            )

            axes_names = [
                stack_suburi.split("/")[-1],
                *grid_names,
            ]

            # Get fit result datasets (inputs) and output dataset information
            input_datasets = list()
            output_info = list()
            output_grps = list()
            for group_name in reversed(list(nxdata_groups)):
                input_grp = xrf_results_grp[group_name]

                output_grp = nexus_utils.create_nxdata(regrid_results, group_name)
                output_grps.append(output_grp)

                signals = list()
                for dset_name, dset in input_grp.items():
                    if not hdf5.is_dataset(dset) or dset_name in axes_names:
                        continue
                    key = group_name, dset_name
                    input_datasets.append(dset)
                    output_info.append((output_grp, group_name, dset_name))
                    signals.append(dset_name)

                nexus_utils.set_nxdata_signals(output_grp, signals=signals)

            # NXdata signals
            nscans = len(bliss_scan_uris)
            output_datasets = dict()
            stack_axis_data = list()
            for scan_index, (bliss_scan_uri, *input_data) in enumerate(
                zip(bliss_scan_uris, *input_datasets)
            ):
                stack_axis_data.append(
                    positioners_utils.get_position_data(bliss_scan_uri, stack_suburi)
                )
                for (output_grp, group_name, dset_name), data in zip(
                    output_info, input_data
                ):
                    data = interpolator.regrid(data)
                    key = group_name, dset_name
                    dset = output_datasets.get(key)
                    if dset is None:
                        stack_shape = (nscans,) + data.shape
                        dset = output_grp.create_dataset(
                            dset_name, shape=stack_shape, dtype=data.dtype
                        )
                        output_datasets[key] = dset
                    dset[scan_index] = data
            stack_axis_values, stack_axis_units = zip(*stack_axis_data)
            stack_axis_units = list(set(stack_axis_units))[0]

            # NXdata axes
            axes_data = [stack_axis_values] + interpolator.grid_axes
            _axes_units = [stack_axis_units] + [
                interpolator.units
            ] * interpolator.grid_ndim
            title_parts = list()
            for iaxis, (axisname, arr, units) in enumerate(
                zip(axes_names, axes_data, _axes_units)
            ):
                title_parts.append(
                    (
                        axisname,
                        len(arr),
                        round_to_significant(abs(arr[1] - arr[0])),
                        units,
                    )
                )
                for output_grp in output_grps:
                    dset = output_grp.create_dataset(axisname, data=arr)
                    if units is not None:
                        dset.attrs["units"] = units
                        dset.attrs["long_name"] = f"{axisname} ({units})"
                    output_grp.create_dataset(f"{axisname}_indices", data=iaxis)
                    output_grp.attrs["axes"] = axes_names

            interpolator.save_coordinates_as_nxdata(regrid_results)

            title = [
                f"{axisname} ({size} x {resolution} {units})"
                for axisname, size, resolution, units in title_parts
            ]
            regrid_results["title"] = " x ".join(title)

    def _get_position_suburis(
        self,
        bliss_scan_uris: Sequence[str],
        ignore_positioners: Optional[Sequence[str]] = None,
    ) -> List[str]:
        positioners = self.inputs.positioners
        if not positioners:
            return positioners_utils.get_scan_position_suburis(
                bliss_scan_uris[0], ignore_positioners=ignore_positioners
            )
        if isinstance(positioners, str):
            positioners = [positioners]
        template = self.inputs.positioner_uri_template
        return [template.format(s) for s in positioners]

    def _get_stack_suburi(self, bliss_scan_uris: Sequence[str]) -> str:
        stack_positioner = self.inputs.stack_positioner
        if stack_positioner:
            template = self.inputs.positioner_uri_template
            return template.format(stack_positioner)
        suburi = get_primary_beam_energy_suburi(bliss_scan_uris[0])
        if not suburi:
            raise RuntimeError("Cannot find energy positioner")
        return suburi



import contextlib
import importlib.metadata
import json
import logging
from datetime import datetime
from typing import Any
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import h5py
import numpy
from ewoksdata.data import nexus
from silx.io import h5py_utils

from ..io import hdf5

version = importlib.metadata.version("ewoksfluo")

logger = logging.getLogger(__name__)


def create_nxdata(
    parent: Union[h5py.File, h5py.Group],
    name: str,
    signal: Optional[str] = None,
) -> h5py.Group:
    nxdata = parent.create_group(name)
    nxdata.attrs["NX_class"] = "NXdata"
    if signal:
        nxdata.attrs["signal"] = signal
    nexus.select_default_plot(nxdata)
    return nxdata


def set_nxdata_signals(nxdata: hdf5.GroupType, signals: Sequence[str]):
    nxdata.attrs["signal"] = signals[0]
    if len(signals) > 1:
        nxdata.attrs["auxiliary_signals"] = signals[1:]


def now() -> str:
    """NeXus-compliant format of the current time"""
    return datetime.now().astimezone().isoformat()


@contextlib.contextmanager
def save_in_ewoks_process(
    output_root_uri: str,
    start_time: str,
    process_config: Dict[str, Any],
    default_levels=("results", "process"),
    **kw,
) -> Generator[Tuple[h5py.Group, bool], None, None]:
    with nexus.create_nexus_group(
        output_root_uri, default_levels=default_levels, **kw
    ) as (
        process_group,
        already_existed,
    ):
        if already_existed:
            logger.warning(
                "%s::%s already exists", process_group.file.filename, process_group.name
            )
            yield process_group, already_existed
        else:
            entry_name = process_group.name.split("/")[1]
            entry_group = process_group.file[entry_name]
            if "start_time" not in entry_group:
                entry_group["start_time"] = start_time

            try:
                process_group.attrs["NX_class"] = "NXprocess"
                process_group["program"] = "ewoksfluo"
                process_group["version"] = version
                config_group = process_group.create_group("configuration")
                config_group.attrs["NX_class"] = "NXnote"
                config_group.create_dataset(
                    "data", data=json.dumps(process_config, cls=NumpyEncoder)
                )
                config_group.create_dataset("date", data=now())
                config_group.create_dataset("type", data="application/json")

                yield process_group, already_existed
            finally:
                if "end_time" in entry_group:
                    entry_group["end_time"][()] = now()
                else:
                    entry_group["end_time"] = now()


@contextlib.contextmanager
def save_in_ewoks_subprocess(
    output_root_uri: str,
    start_time: str,
    process_config: Dict[str, Any],
    collection_name: str = "results",
    **kw,
) -> Generator[Tuple[h5py.Group, bool], None, None]:
    with save_in_ewoks_process(output_root_uri, start_time, process_config, **kw) as (
        process_group,
        already_existed,
    ):
        if already_existed:
            results = process_group[collection_name]
        else:
            results = process_group.create_group(collection_name)
            results.attrs["NX_class"] = "NXcollection"
        yield results, already_existed


class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, (numpy.generic, numpy.ndarray)):
            return obj.tolist()
        return super().default(obj)


@h5py_utils.retry(retry_timeout=20)
def wait_scan_finished(bliss_scan_uri: str) -> None:
    input_file, scan_h5path = hdf5.split_h5uri(bliss_scan_uri)
    with h5py_utils.File(input_file) as fh:
        parts = [p for p in scan_h5path.split("/") if p]
        if parts:
            scan = parts[0]
            _ = fh[scan]["end_time"][()]



from typing import Dict
from typing import List
from typing import Optional
from typing import Union

from ewokscore import Task
from ewokscore.model import BaseOutputModel
from pydantic import Field

from ...xrffit.pymca_config import PyMcaXrfConfiguration
from ..models import FluoInputs
from ..positioner_utils import get_primary_beam_energy_value
from .execute import fit_multi
from .execute import fit_single


class FitSingleScanSingleDetectorInputs(FluoInputs):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )

    detector_name: str = Field(
        description="NXdetector group name.",
        min_length=1,
        examples=["fx_nano_det0"],
    )

    output_root_uri: str = Field(
        description="Output root URI.",
        examples=[
            "/results/dataset.h5",
            "/results/dataset.h5::/1.1",
            "/results/dataset.h5::/1.1/fit",
        ],
    )
    config: Union[str, PyMcaXrfConfiguration] = Field(
        description="PyMca configuration",
        examples=["/config/configuration.cfg", {"peaks": {}, "fit": {}}],
    )

    xrf_spectra_uri_template: str = Field(
        default="instrument/{}/data",
        description="HDF5 group URI template of an XRF detector relative to `bliss_scan_uri`.",
    )

    process_uri_template: str = Field(default="fit/{}")
    energy_name: Optional[str] = None
    energy_uri_template: str = Field(default="instrument/positioners_start/{}")
    quantification: Union[Dict[str, float], bool, None] = None
    energy_multiplier: Optional[float] = None
    fast_fitting: bool = Field(default=True)
    diagnostics: bool = Field(default=False)
    figuresofmerit: Optional[bool] = None


class FitSingleScanSingleDetectorOutputs(BaseOutputModel):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )
    output_root_uri: str = Field(
        description="Output NXentry URI.", examples=["/results/dataset.h5::/1.1"]
    )
    xrf_results_uri: str = Field(
        description="Output NXprocess results URI.",
        examples=["/results/dataset.h5::/1.1/fit/mca0/results"],
    )
    detector_name: str = Field(description="Name of the fitted detector.")


class FitSingleScanSingleDetector(
    Task,
    input_model=FitSingleScanSingleDetectorInputs,
    output_model=FitSingleScanSingleDetectorOutputs,
):
    """XRF fit of one scan with one detector"""

    def run(self):
        params = self.get_input_values()
        params["pymca_configuration"] = params.pop("config")
        _parse_energy(params)
        xrf_results_uri, output_root_uri = fit_single(**params)
        self.outputs.xrf_results_uri = xrf_results_uri
        self.outputs.bliss_scan_uri = params["bliss_scan_uri"]
        self.outputs.detector_name = params["detector_name"]
        self.outputs.output_root_uri = output_root_uri


class FitSingleScanMultiDetectorInputs(FluoInputs):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )

    detector_names: List[str] = Field(
        description="NXdetector group name.",
        min_length=1,
        examples=[["fx_nano_det0", "fx_nano_det1"]],
    )

    output_root_uri: str = Field(
        description="Output root URI.",
        examples=[
            "/results/dataset.h5",
            "/results/dataset.h5::/1.1",
            "/results/dataset.h5::/1.1/fit",
        ],
    )
    configs: List[Union[str, PyMcaXrfConfiguration]] = Field(
        description="PyMca configuration",
        examples=[
            ["/config/det0.cfg", "/config/det1.cfg"],
            [{"peaks": {}, "fit": {}}, {"peaks": {}, "fit": {}}],
        ],
    )

    xrf_spectra_uri_template: str = Field(
        default="instrument/{}/data",
        description="HDF5 group URI template of an XRF detector relative to `bliss_scan_uri`.",
    )

    process_uri_template: str = Field(default="fit/{}")
    energy_name: Optional[str] = None
    energy_uri_template: str = Field(default="instrument/positioners_start/{}")
    quantification: Union[Dict[str, float], bool, None] = None
    energy_multiplier: Optional[float] = None
    fast_fitting: bool = Field(default=True)
    diagnostics: bool = Field(default=False)
    figuresofmerit: Optional[bool] = None


class FitSingleScanMultiDetectorOutputs(BaseOutputModel):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )
    output_root_uri: str = Field(
        description="Output NXentry URI.", examples=["/results/dataset.h5::/1.1"]
    )
    xrf_results_uris: List[str] = Field(
        description="Output NXprocess results URI.",
        examples=[
            "/results/dataset.h5::/1.1/fit/fx_nano_det0/results",
            "/results/dataset.h5::/1.1/fit/fx_nano_det1/results",
        ],
    )
    detector_names: List[str] = Field(description="Names of the fitted detectors.")


class FitSingleScanMultiDetector(
    Task,
    input_model=FitSingleScanMultiDetectorInputs,
    output_model=FitSingleScanMultiDetectorOutputs,
):
    """XRF fit of one scan with multiple detectors"""

    def run(self):
        params = self.get_input_values()
        params["pymca_configurations"] = params.pop("configs")
        bliss_scan_uri = params.pop("bliss_scan_uri")
        params["bliss_scan_uris"] = [bliss_scan_uri]
        _parse_energies(params)
        xrf_results_uris, output_root_uri = fit_multi(**params)
        self.outputs.xrf_results_uris = xrf_results_uris
        self.outputs.bliss_scan_uri = bliss_scan_uri
        self.outputs.detector_names = params["detector_names"]
        self.outputs.output_root_uri = output_root_uri


class FitStackSingleDetectorInputs(FluoInputs):
    bliss_scan_uris: List[str] = Field(
        description="Bliss scan URI's.",
        examples=[["/data/dataset.h5::/1.1", "/data/dataset.h5::/2.1"]],
    )

    detector_name: str = Field(
        description="NXdetector group name.",
        min_length=1,
        examples=["fx_nano_det0"],
    )

    output_root_uri: str = Field(
        description="Output root URI.",
        examples=[
            "/results/dataset.h5",
            "/results/dataset.h5::/1.1",
            "/results/dataset.h5::/1.1/fit",
        ],
    )
    config: Union[str, PyMcaXrfConfiguration] = Field(
        description="PyMca configuration",
        examples=["/config/configuration.cfg", {"peaks": {}, "fit": {}}],
    )

    xrf_spectra_uri_template: str = Field(
        default="instrument/{}/data",
        description="HDF5 group URI template of an XRF detector relative to `bliss_scan_uri`.",
    )

    process_uri_template: str = Field(default="fit/{}")
    energy_name: Optional[str] = None
    energy_uri_template: str = Field(default="instrument/positioners_start/{}")
    quantification: Union[Dict[str, float], bool, None] = None
    energy_multiplier: Optional[float] = None
    fast_fitting: bool = Field(default=True)
    diagnostics: bool = Field(default=False)
    figuresofmerit: Optional[bool] = None


class FitStackSingleDetectorOutputs(BaseOutputModel):
    bliss_scan_uris: List[str] = Field(
        description="Bliss scan URI's.",
        examples=[["/data/dataset.h5::/1.1", "/data/dataset.h5::/2.1"]],
    )
    output_root_uri: str = Field(
        description="Output NXentry URI.", examples=["/results/dataset.h5::/1.1"]
    )
    xrf_results_uri: str = Field(
        description="Output NXprocess results URI.",
        examples=["/results/dataset.h5::/1.1/fit/fx_nano_det0/results"],
    )
    detector_name: str = Field(description="Name of the fitted detector.")


class FitStackSingleDetector(
    Task,
    input_model=FitStackSingleDetectorInputs,
    output_model=FitStackSingleDetectorOutputs,
):
    """XRF fit of a stack of identical scan with one detector"""

    def run(self):
        params = self.get_input_values()
        detector_name = params.pop("detector_name")
        params["detector_names"] = [detector_name]
        params["pymca_configurations"] = [params.pop("config")]
        _parse_energies(params)
        xrf_results_uris, output_root_uri = fit_multi(**params)
        self.outputs.xrf_results_uri = xrf_results_uris[0]
        self.outputs.bliss_scan_uris = params["bliss_scan_uris"]
        self.outputs.detector_name = detector_name
        self.outputs.output_root_uri = output_root_uri


class FitStackMultiDetectorInputs(FluoInputs):
    bliss_scan_uris: List[str] = Field(
        description="Bliss scan URI's.",
        examples=[["/data/dataset.h5::/1.1", "/data/dataset.h5::/2.1"]],
    )

    detector_names: List[str] = Field(
        description="NXdetector group name.",
        min_length=1,
        examples=[["fx_nano_det0", "fx_nano_det1"]],
    )

    output_root_uri: str = Field(
        description="Output root URI.",
        examples=[
            "/results/dataset.h5",
            "/results/dataset.h5::/1.1",
            "/results/dataset.h5::/1.1/fit",
        ],
    )
    configs: List[Union[str, PyMcaXrfConfiguration]] = Field(
        description="PyMca configuration",
        examples=[
            ["/config/det0.cfg", "/config/det1.cfg"],
            [{"peaks": {}, "fit": {}}, {"peaks": {}, "fit": {}}],
        ],
    )

    xrf_spectra_uri_template: str = Field(
        default="instrument/{}/data",
        description="HDF5 group URI template of an XRF detector relative to `bliss_scan_uri`.",
    )

    process_uri_template: str = Field(default="fit/{}")
    energy_name: Optional[str] = None
    energy_uri_template: str = Field(default="instrument/positioners_start/{}")
    quantification: Union[Dict[str, float], bool, None] = None
    energy_multiplier: Optional[float] = None
    fast_fitting: bool = Field(default=True)
    diagnostics: bool = Field(default=False)
    figuresofmerit: Optional[bool] = None


class FitStackMultiDetectorOutputs(BaseOutputModel):
    bliss_scan_uris: List[str] = Field(
        description="Bliss scan URI's.",
        examples=[["/data/dataset.h5::/1.1", "/data/dataset.h5::/2.1"]],
    )
    output_root_uri: str = Field(
        description="Output NXentry URI.", examples=["/results/dataset.h5::/1.1"]
    )
    xrf_results_uris: List[str] = Field(
        description="Output NXprocess results URI.",
        examples=[
            "/results/dataset.h5::/1.1/fit/fx_nano_det0/results",
            "/results/dataset.h5::/1.1/fit/fx_nano_det1/results",
        ],
    )
    detector_names: List[str] = Field(description="Names of the fitted detectors.")


class FitStackMultiDetector(
    Task,
    input_model=FitStackMultiDetectorInputs,
    output_model=FitStackMultiDetectorOutputs,
):
    """XRF fit of a stack of identical scan with multiple detectors"""

    def run(self):
        params = self.get_input_values()
        params["pymca_configurations"] = params.pop("configs")
        _parse_energies(params)
        xrf_results_uris, output_root_uri = fit_multi(**params)
        self.outputs.xrf_results_uris = xrf_results_uris
        self.outputs.bliss_scan_uris = params["bliss_scan_uris"]
        self.outputs.detector_names = params["detector_names"]
        self.outputs.output_root_uri = output_root_uri


def _parse_energy(params: dict) -> None:
    energy_name = params.pop("energy_name", None)
    energy_uri_template = params.pop("energy_uri_template", None)
    params["energy"] = get_primary_beam_energy_value(
        params["bliss_scan_uri"],
        energy_name=energy_name,
        energy_uri_template=energy_uri_template,
        search_on_units=False,
    )


def _parse_energies(params: dict) -> None:
    energy_name = params.pop("energy_name", None)
    energy_uri_template = params.pop("energy_uri_template", None)
    params["energies"] = [
        get_primary_beam_energy_value(
            bliss_scan_uri,
            energy_name=energy_name,
            energy_uri_template=energy_uri_template,
            search_on_units=False,
        )
        for bliss_scan_uri in params["bliss_scan_uris"]
    ]



import psutil
import pytest
from ewoksorange.tests.conftest import ewoks_orange_canvas  # noqa F401
from ewoksorange.tests.conftest import qtapp  # noqa F401
from ewoksorange.tests.conftest import raw_ewoks_orange_canvas  # noqa F401

from .. import resource_utils


@pytest.fixture(params=["normal", "slurm"])
def system_4cpus_8gb(request, monkeypatch) -> None:
    # Force known CPU count of 4 and memory of 8 GB

    class FakeProcess:
        def cpu_affinity(self):
            return [0, 1, 2, 3]  # Simulate 4 CPUs

    monkeypatch.setattr(psutil, "Process", lambda: FakeProcess())

    if request.param == "normal":
        monkeypatch.setattr(
            resource_utils, "_get_available_memory", lambda: 8 * 1024**3
        )
    elif request.param == "slurm":

        class SubprocessResponse:
            def __init__(self, stdout):
                self.stdout = stdout
                self.returncode = 0

        def mock_subprocess_run(cmd, **kwargs):
            if "sacct" in cmd:
                return SubprocessResponse("16G\n")  # 16 GB requested
            elif "sstat" in cmd:
                return SubprocessResponse("8G\n")  # 8 GB used
            return SubprocessResponse("")

        monkeypatch.setenv("SLURM_JOB_ID", "12345")
        monkeypatch.setattr(resource_utils.subprocess, "run", mock_subprocess_run)
    else:
        raise ValueError(f"{request.param}")

    # Verify that the patching works as expected
    cpus = resource_utils.get_available_cpus(exclude_current=True)
    assert cpus == 3
    mem = resource_utils._get_available_memory()
    assert mem == 8 * 1024**3



from typing import NamedTuple
from typing import Optional
from typing import Union

import numpy
from typing_extensions import TypeGuard


class StopMsg(NamedTuple):
    sendid: int
    cmd: str = "STOP"


class GroupMsg(NamedTuple):
    sendid: int
    destinationid: int
    group: str
    data: dict
    cmd: str = "GROUP"


class DatasetMsg(NamedTuple):
    sendid: int
    destinationid: int
    group: str
    name: str
    npoints: int
    attrs: dict
    nscans: Optional[int]
    cmd: str = "DATASET"


class DataMsg(NamedTuple):
    sendid: int
    destinationid: int
    group: str
    name: str
    value: numpy.ndarray
    scan_index: Optional[int]
    cmd: str = "DATA"


QueueMsg = Union[StopMsg, GroupMsg, DatasetMsg, DataMsg]


def is_StopMsg(msg: QueueMsg) -> TypeGuard[StopMsg]:
    return msg.cmd == "STOP"


def is_GroupMsg(msg: QueueMsg) -> TypeGuard[GroupMsg]:
    return msg.cmd == "GROUP"


def is_DatasetMsg(msg: QueueMsg) -> TypeGuard[DatasetMsg]:
    return msg.cmd == "DATASET"


def is_DataMsg(msg: QueueMsg) -> TypeGuard[DataMsg]:
    return msg.cmd == "DATA"



import itertools
import logging
from typing import Any
from typing import Dict
from typing import Generator
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy
from numpy.typing import DTypeLike

from ...io import hdf5
from ...resource_utils import array_chunk_size
from ...resource_utils import log_required_memory
from .. import xrf_results
from ..math import eval_hdf5_expression
from ..math import format_expression_template

_logger = logging.getLogger(__name__)


def detector_weight_iterator(
    weight_uri_root: str, weight_expressions: Sequence[str]
) -> Generator[numpy.ndarray, None, None]:
    """
    :param weight_uri_root: HDF5 root group under which the detector weight datasets can be found.
    :param weight_expressions: Arithmetic expression for each detector to calculated the weight for addition from HDF5 datasets.
    """
    for weight_expression in weight_expressions:
        yield eval_hdf5_expression(weight_uri_root, weight_expression)


def detector_weight_iterator_stack(
    weight_uri_roots: Sequence[str], weight_expressions: Sequence[str]
) -> Generator[numpy.ndarray, None, None]:
    """
    :param weight_uri_roots: HDF5 root group under which the detector weight datasets can be found.
    :param weight_expressions: Arithmetic expression for each detector to calculated the weight for addition from HDF5 datasets.
    """
    for weight_expression in weight_expressions:
        detector_weight = [
            eval_hdf5_expression(weight_uri_root, weight_expression)
            for weight_uri_root in weight_uri_roots
        ]
        yield numpy.stack(detector_weight, axis=0)


def save_summed_xrf_results(
    xrf_results_uris: Sequence[str],
    detector_weights: Iterator[numpy.ndarray],
    output_root_uri: str,
    process_config: Dict[str, Any],
) -> str:
    """
    :param xrf_results_uris: HDF5 group for each detector that contains the "parameters", "uncertainties" and "massfractions" groups.
    :param detector_weights: Weights for each detector.
    :param output_root_uri: Root HDF5 URL under which to save the results as NXprocess
    :returns: URI to summed fit result group
    """
    parameters, uncertainties, massfractions = compute_summed_xrf_results(
        xrf_results_uris, detector_weights
    )
    return xrf_results.save_xrf_results(
        output_root_uri,
        "sum",
        process_config,
        parameters,
        uncertainties,
        massfractions,
    )


def compute_summed_xrf_results(
    xrf_results_uris: Sequence[str], detector_weights: Iterator[numpy.ndarray]
) -> Tuple[
    Dict[str, numpy.ndarray], Dict[str, numpy.ndarray], Dict[str, numpy.ndarray]
]:
    r"""Compute the weighted sum of the peak areas, associated uncertainties and mass fractions for several detectors.

    For elemental peak areas

    .. math::

        A(\mathrm{Fe}) = \sum_i{\left[ W_i A_i(\mathrm{Fe}) \right] }

    For their uncertainties

    .. math::

        \sigma_{A}(\mathrm{Fe}) = \sqrt{ \sum_i\left[ W_i^2 \sigma_{A_i}^2(\mathrm{Fe}) \right] }

    For elemental mass fractions, in addition to the detector weight we also weight for the peak area

    .. math::

        M(\mathrm{Fe}) = \sum_i{\left[ W_i M_i(\mathrm{Fe}) \frac{W_i A_i(\mathrm{Fe})}{A(\mathrm{Fe})} \right] }

    The variable :math:`W_i` is the weight for detector :math:`i` which is typically the inverse of the live time.

    :param xrf_results_uris: HDF5 group for each detector that contains the "parameters", "uncertainties" and "massfractions" groups.
    :param detector_weights: Weights for each detector.
    :returns: summed peak areas, associated uncertainties, averaged weight fractions
    """
    summed_parameters = {}
    summed_variances = {}
    averaged_massfractions = {}

    for xrf_results_uri, detector_weight in itertools.zip_longest(
        xrf_results_uris, detector_weights, fillvalue=None
    ):
        if xrf_results_uri is None or detector_weight is None:
            raise ValueError(
                "The number of arithmetic expressions but be equal to the number of detectors"
            )

        fit_filename, fit_h5path = xrf_results_uri.split("::")

        with hdf5.FileReadAccess(fit_filename) as h5file:
            xrf_results_group = h5file[fit_h5path]
            assert hdf5.is_group(xrf_results_group)

            # Sum the peak areas and average mass fractions
            param_group = xrf_results_group["parameters"]
            assert hdf5.is_group(param_group)
            massfrac_group = xrf_results_group.get("massfractions", dict())
            for dset_name, dset in param_group.items():
                if not xrf_results.is_peak_area(dset):
                    continue

                wparam_value = dset[()] * detector_weight
                if dset_name in summed_parameters:
                    summed_parameters[dset_name] += wparam_value
                else:
                    summed_parameters[dset_name] = wparam_value

                if dset_name not in massfrac_group:
                    continue

                wmassfrac_value = massfrac_group[dset_name][()] * detector_weight
                wmassfrac_num = wparam_value * wmassfrac_value
                if dset_name in averaged_massfractions:
                    averaged_massfractions[dset_name] += wmassfrac_num
                else:
                    averaged_massfractions[dset_name] = wmassfrac_num

            # Propagate error on peak areas
            uncertainties_group = xrf_results_group["uncertainties"]
            assert hdf5.is_group(uncertainties_group)
            for dset_name, dset in uncertainties_group.items():
                if not hdf5.is_dataset(dset):
                    continue
                wvar_value = dset[()] ** 2 * detector_weight**2
                if dset_name in summed_variances:
                    summed_variances[dset_name] += wvar_value
                else:
                    summed_variances[dset_name] = wvar_value

    summed_uncertainties = {k: numpy.sqrt(v) for k, v in summed_variances.items()}

    if averaged_massfractions:
        averaged_massfractions = {
            k: v / summed_parameters[k] for k, v in averaged_massfractions.items()
        }

    return summed_parameters, summed_uncertainties, averaged_massfractions


def sum_spectra_from_hdf5(
    bliss_scan_uri: str,
    xrf_spectra_uri_template: str,
    detector_normalization_template: Optional[str],
    detector_names: List[str],
) -> Optional[numpy.ndarray]:
    r"""Add spectra from multiple detectors after normalizing the spectra to a common live time.

    .. math::

        MCA(\mathrm{Fe}) = \sum_i{\left[ W_i MCA_i \right] }

    :param bliss_scan_uri: BLISS scan URI. For example `"/path/to/data.g5::/1.1"`.
    :param xrf_spectra_uri_template: XRF spectra path with respect to the BLISS scan URI. For example `"instrument/{}/data"`.
    :param detector_normalization_template: XRF spectra normalization template with variables
                                            with respect to the BLISS scan URI. For example `"0.1/<instrument/{}/live_time>"`.
    :param detector_names: List of names to be used in the templates.
    :returns: summed XRF spectra
    """
    input_file, scan_h5path = hdf5.split_h5uri(bliss_scan_uri)
    with hdf5.FileReadAccess(input_file) as h5file:
        sum_spectra = None
        chunk_size = 1
        scan_group = h5file[scan_h5path]

        for detector_name in detector_names:
            # Check dataset with XRF spectra
            xrf_spectra_uri = xrf_spectra_uri_template.format(detector_name)
            xrf_spectra_dataset = scan_group[xrf_spectra_uri]
            if not hdf5.is_dataset(xrf_spectra_dataset):
                raise ValueError(f"{xrf_spectra_uri!r} is not an HDF5 dataset")
            if not xrf_spectra_dataset.ndim == 2:
                raise ValueError(f"{xrf_spectra_uri!r} is not 2D")
            shape = xrf_spectra_dataset.shape
            dtype = xrf_spectra_dataset.dtype
            if xrf_spectra_dataset.dtype == numpy.float64:
                weight_dtype = numpy.float64
            else:
                weight_dtype = numpy.float32

            # Calculate the detector weights for the sum
            weights = _get_detector_weights(
                bliss_scan_uri,
                detector_name,
                detector_normalization_template,
                weight_dtype,
            )
            if weights is None:
                final_dtype = dtype
            else:
                final_dtype = weight_dtype
                if weights.ndim > 1:
                    raise ValueError(
                        f"Weights for detector {detector_name!r} is not a scalar or 1D"
                    )
                if weights.size != shape[0]:
                    raise ValueError(
                        f"Number of detector weights and number of spectra for {detector_name!r} are not the same."
                    )
                weights = weights.astype(final_dtype)

            # Prepare array to hold the sum
            if sum_spectra is None:
                if weights is None:
                    # Memory: 1 array
                    multiplier = 1
                else:
                    # Memory: 1 array and 1 cast array
                    multiplier = 2

                log_required_memory(
                    "initialize sum spectrum",
                    shape,
                    dtype,
                    multiplier=multiplier + 0.1,
                )
                sum_spectra = xrf_spectra_dataset[()]
                if weights is not None:
                    sum_spectra = sum_spectra.astype(final_dtype)
                    sum_spectra *= weights.reshape(-1, 1)

                chunk_size = array_chunk_size(shape, dtype, multiplier=multiplier + 0.1)
                log_required_memory(
                    "sum XRF spectra with chunking",
                    shape,
                    dtype,
                    multiplier=multiplier * chunk_size / shape[0],
                )
                continue

            # Add spectra
            if weights is None:
                for i in range(0, shape[0], chunk_size):
                    chunk = xrf_spectra_dataset[i : i + chunk_size]
                    sum_spectra[i : i + chunk_size] += chunk
            else:
                for i in range(0, shape[0], chunk_size):
                    chunk = xrf_spectra_dataset[i : i + chunk_size]
                    chunk = chunk.astype(final_dtype)
                    w_chunk = weights[i : i + chunk_size].reshape(-1, 1)
                    chunk *= w_chunk
                    sum_spectra[i : i + chunk_size] += chunk

    return sum_spectra


def _get_detector_weights(
    bliss_scan_uri: str,
    detector_name: str,
    detector_normalization_template: Optional[str],
    cast_dtype: DTypeLike,
) -> Union[numpy.ndarray, None]:
    if detector_normalization_template is None:
        _logger.warning(
            "Add detector %r to sum spectra WITHOUT weights",
            detector_name,
        )
        return

    weight_expression = format_expression_template(
        detector_normalization_template, detector_name
    )
    _logger.info(
        "Add detector %r to sum spectra with weights %s",
        detector_name,
        weight_expression,
    )
    weights = eval_hdf5_expression(bliss_scan_uri, weight_expression)
    return numpy.asarray(weights, dtype=cast_dtype)



"""Read data from SPEC files"""

import os
import re
from glob import glob
from typing import Dict
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple

import numpy
from silx.io.specfile import SpecFile as _SpecFile

from . import types
from .zap_utils import get_zap_positioners
from .zap_utils import iter_zap_data
from .zap_utils import rename_counter


class SpecFile:
    def __init__(self, filename: str):
        self._filename = filename
        self._specfile = None

    def __enter__(self):
        self._specfile = _SpecFile(self._filename)
        return self

    def __exit__(self, *args):
        self.close()

    def close(self):
        self._specfile.close()
        self._specfile = None

    def get_scans(self) -> List[Tuple[int, int]]:
        return [tuple(map(int, k.split("."))) for k in self._specfile.keys()]

    def get_number_of_subsubscans(self, scannb: int, subscan: int = 1) -> int:
        scan = self._specfile[f"{scannb}.{subscan}"]
        title = self.get_title(scannb, subscan)
        if "fXAS_scan" in title:
            return scan.data.shape[1]
        return 1

    def iter_data(
        self,
        scannb: int,
        subscan: int = 1,
        subsubscan: int = 1,
        exclude_mca_detectors: Sequence[int] = tuple(),
    ) -> Iterator[types.ScanData]:
        scan = self._specfile[f"{scannb}.{subscan}"]
        header = scan.scan_header_dict

        motor_names = [rename_counter(s) for s in scan.motor_names]
        labels = [rename_counter(s) for s in scan.labels]
        spec_data = scan.data
        title = self.get_title(scannb, subscan)

        if "fXAS_scan" in title:
            zap_info = _zap_info_from_fluoxas_scan(
                header, spec_data.shape[1], subsubscan
            )
            if zap_info:
                yield from iter_zap_data(zap_info, exclude_mca_detectors, motor_names)
        else:
            for name, data in zip(labels, spec_data):
                if name in motor_names or name in header["S"]:
                    detector_type = "positioner"
                    data_name = "value"
                else:
                    detector_type = ""
                    data_name = "data"
                yield types.ScanData(
                    group=name,
                    name=data_name,
                    detector_type=detector_type,
                    data=numpy.array(data),
                    local_alias="",
                    global_alias=name,
                )
            yield from _iter_motors_from_title(title, labels)
            zap_info = _zap_info_from_scan(header)
            if zap_info:
                yield from iter_zap_data(zap_info, exclude_mca_detectors, motor_names)

    def get_positioners(
        self, scannb: int, subscan: int = 1, subsubscan: int = 1
    ) -> Dict[str, float]:
        scan = self._specfile[f"{scannb}.{subscan}"]
        motor_names = [rename_counter(s) for s in scan.motor_names]
        title = self.get_title(scannb, subscan)
        if "fXAS_scan" in title:
            zap_info = _zap_info_from_fluoxas_scan(
                scan.scan_header_dict, scan.data.shape[1], subsubscan
            )
        else:
            zap_info = _zap_info_from_scan(scan.scan_header_dict)
        if zap_info:
            positioners = get_zap_positioners(zap_info)
        else:
            positioners = dict()
        positioners.update((zip(motor_names, scan.motor_positions)))
        return positioners

    def get_title(self, scannb: int, subscan: int = 1) -> str:
        scan = self._specfile[f"{scannb}.{subscan}"]
        title = re.split(r"\s", scan.scan_header_dict["S"])
        return " ".join([s for s in title[1:] if s]).strip()


def _zap_info_from_scan(header: Dict[str, str]) -> Optional[types.ZapInfo]:
    xia_keys = {
        "DIRECTORY": ("directory", str),
        "RADIX": ("radix", str),
        "ZAP SCAN NUMBER": ("scannb", int),
        "ZAP IMAGE NUMBER": ("zapnb", int),
    }
    info = _xia_info_from_header(header, xia_keys)
    if info is None:
        return
    return types.ZapInfo(**info)


def _zap_info_from_fluoxas_scan(
    header: Dict[str, str], npoints: int, subsubscan: int
) -> Optional[types.ZapInfo]:
    xia_keys = {
        "DIRECTORY": ("directory", str),
    }
    info = _xia_info_from_header(header, xia_keys)
    if info is None:
        return
    directory, radix = os.path.split(info["directory"])
    radix_prefix, _, nb_stop = radix.rpartition("_")
    nb_stop = int(nb_stop)
    nb_start = nb_stop - npoints
    zapnb = 0
    nb = nb_start + subsubscan
    radix = f"{radix_prefix}_{nb:04d}"
    zap_directory = os.path.join(directory, radix)

    pattern = os.path.join(zap_directory, f"{radix}_xiast_*_{zapnb:04d}_0000.edf")
    files = glob(pattern)
    if len(files) != 1:
        return
    repattern = os.path.join(
        zap_directory, f"{radix}_xiast_([0-9]+)_{zapnb:04d}_0000.edf"
    )
    m = re.match(repattern, files[0])
    if not m:
        return
    scannb = int(m.groups()[0])
    return types.ZapInfo(
        directory=zap_directory, radix=radix, scannb=scannb, zapnb=zapnb
    )


def _xia_info_from_header(header: dict, xia_keys: dict) -> Optional[dict]:
    info = dict()
    for s in header.get("C", "").split("\n"):
        tmp = s.split(":")
        if len(tmp) != 2:
            continue
        key, value = [s.strip() for s in tmp]
        key, deserialize = xia_keys.get(key, (None, None))
        if key:
            info[key] = deserialize(value)
    if len(info) != len(xia_keys):
        return
    return info


def _iter_motors_from_title(
    title: str, exclude: Sequence[str] = tuple()
) -> Iterator[types.ScanData]:
    title = [s for s in re.split(r"\s", title) if s]
    if title[0] == "zapline":
        try:
            cmd, motfast, startfast, endfast, npixelsfast, time = title
        except ValueError:
            return
        if motfast not in exclude:
            data = _contspace(startfast, endfast, npixelsfast)
            yield types.ScanData(
                group=motfast,
                name="value",
                detector_type="positioner",
                data=data,
                local_alias="",
                global_alias=motfast,
            )
    elif title[0] == "zapimage":
        try:
            (
                cmd,
                motfast,
                startfast,
                endfast,
                npixelsfast,
                time,
                motslow,
                startslow,
                endslow,
                nstepsslow,
                zero,
            ) = title
        except ValueError:
            return
        if motfast not in exclude:
            data = numpy.tile(
                _contspace(startfast, endfast, npixelsfast),
                int(nstepsslow) + 1,
            )
            yield types.ScanData(
                group=motfast,
                name="value",
                detector_type="positioner",
                data=data,
                local_alias="",
                global_alias=motfast,
            )
        if motslow not in exclude:
            data = numpy.repeat(
                _stepspace(startslow, endslow, nstepsslow),
                int(npixelsfast),
            )
            yield types.ScanData(
                group=motslow,
                name="value",
                detector_type="positioner",
                data=data,
                local_alias="",
                global_alias=motslow,
            )
    elif title[0] == "ascan":
        try:
            cmd, motfast, startfast, endfast, nstepsfast, time = title
        except ValueError:
            return
        if motfast not in exclude:
            data = _stepspace(startfast, endfast, nstepsfast)
            yield types.ScanData(
                group=motfast,
                name="value",
                detector_type="positioner",
                data=data,
                local_alias="",
                global_alias=motfast,
            )


def _contspace(start: str, stop: str, npoints: str) -> numpy.ndarray:
    start = float(start)
    stop = float(stop)
    npoints = int(npoints)
    nsteps = npoints - 1
    half_step = (stop - start) / (2 * nsteps)
    return numpy.linspace(start + half_step, stop - half_step, npoints)


def _stepspace(start: str, stop: str, nsteps: str) -> numpy.ndarray:
    start = float(start)
    stop = float(stop)
    npoints = int(nsteps) + 1
    return numpy.linspace(start, stop, npoints)



from typing import Callable
from typing import List
from typing import Tuple
from typing import Union

import numpy
from PyMca5.PyMcaPhysics.xrf.ClassMcaTheory import ClassMcaTheory
from PyMca5.PyMcaPhysics.xrf.ConcentrationsTool import ConcentrationsTool

from ...pymca_config import PyMcaXrfConfiguration
from ..types import QuantificationParameters


def calculate_massfractions(
    mca_theory: ClassMcaTheory,
    peak_group_areas: numpy.ndarray,
    peak_group_names: List[str],
    configuration: PyMcaXrfConfiguration,
    norm_factors: Union[float, numpy.ndarray],
) -> Tuple[numpy.ndarray, List[numpy.ndarray], List[str], QuantificationParameters]:
    """Calculate mass fractions of the fit of many spectra based on the fit of a reference spectrum and
    fundamental parameters or an internal reference.

    Without internal reference (fundamental parameters):

        mass_fraction(group) = norm_factors * peak_area(group)  / ref_fit_peak_area(group)  * ref_fit_mass_fraction(group)

    With internal reference:

        mass_fraction(group) = peak_area(group) / peak_area(int) * ref_fit_peak_area(int) / ref_fit_peak_area(group) * ref_fit_mass_fraction(group)

    :param mca_theory:
    :param peak_group_areas: shape ``(num_parameters, num_profiles)``
    :param peak_group_names:
    :param configuration:
    :param norm_factors: scalar or shape ``(num_profiles, )`` (only for fundamental parameters)
    :returns: mass fractions with shape ``(num_parameters, num_profiles)``,
              mass fractions per layer with shape ``(num_layers, num_parameters, num_profiles)``,
              mass fraction names,
              internal reference element
    """

    ctool = ConcentrationsTool()
    ctool_config = ctool.configure()
    ctool_config.update(configuration.concentrations.model_dump(mode="python"))

    # Prepare fit results
    usematrix = configuration.concentrations.usematrix
    reference = configuration.concentrations.reference
    select_fitreference = usematrix and reference.upper() == "AUTO"
    configuration = configuration.model_dump(mode="python")  # copy
    if select_fitreference:
        # Select internal reference
        _ = mca_theory.startfit(digest=0, linear=True)
        fitresult = {"result": mca_theory.imagingDigestResult()}
        fitresult["result"]["config"] = configuration
        _, add_info = ctool.processFitResult(
            config=ctool_config,
            fitresult=fitresult,
            elementsfrommatrix=False,
            fluorates=mca_theory._fluoRates,
            addinfo=True,
        )
        for group in fitresult["result"]["groups"]:
            if fitresult["result"][group]["fitarea"] <= 0.0:
                fitresult["result"][group]["fitarea"] = 1.0e-6
        configuration["concentrations"]["reference"] = add_info["ReferenceElement"]
    else:
        fitresult = {"result": {"config": configuration, "groups": peak_group_names}}
        for name in peak_group_names:
            # We just need the ratio between peak area and concentration
            fitresult["result"][name] = {"fitarea": 1.0, "sigmaarea": 1.0}

    # Calculate mass fractions of reference fit
    wresult, add_info = ctool.processFitResult(
        config=ctool_config,
        fitresult=fitresult,
        elementsfrommatrix=False,
        fluorates=mca_theory._fluoRates,
        addinfo=True,
    )
    quantification_parameters = QuantificationParameters(**add_info)

    # Choose calculation mode
    if quantification_parameters.ReferenceElement in ["", None, "None"]:
        # Fundamental parameters

        def _calc_mass_fraction(
            name, peak_area, ref_fit_peak_area, ref_fit_mass_fraction, layer=None
        ):
            return _calc_mass_fraction_fp(
                peak_area, ref_fit_peak_area, ref_fit_mass_fraction, norm_factors
            )

        mass_fractions, mass_fractions_per_layer, mass_fraction_names = (
            _calc_mass_fractions(
                fitresult["result"]["groups"],
                peak_group_areas,
                wresult,
                fitresult,
                _calc_mass_fraction,
            )
        )
    else:
        # Internal reference
        reference_group = (
            quantification_parameters.ReferenceElement
            + " "
            + quantification_parameters.ReferenceTransitions.split()[0]
        )
        idx_int = fitresult["result"]["groups"].index(reference_group)
        name_int = fitresult["result"]["groups"][idx_int]
        peak_area_int = peak_group_areas[idx_int]
        ref_fit_peak_area_int = fitresult["result"][name_int]["fitarea"]

        def _calc_mass_fraction(
            name, peak_area, ref_fit_peak_area, ref_fit_mass_fraction, layer=None
        ):
            if name == name_int:
                if layer is None:
                    return [wresult["mass fraction"][name]] * len(peak_area)
                return [wresult[layer]["mass fraction"][name]] * len(peak_area)

            return _calc_mass_fraction_int(
                peak_area,
                ref_fit_peak_area,
                ref_fit_mass_fraction,
                peak_area_int,
                ref_fit_peak_area_int,
            )

        mass_fractions, mass_fractions_per_layer, mass_fraction_names = (
            _calc_mass_fractions(
                fitresult["result"]["groups"],
                peak_group_areas,
                wresult,
                fitresult,
                _calc_mass_fraction,
            )
        )

    # Convert lists to numpy arrays
    mass_fractions_per_layer = [numpy.asarray(lst) for lst in mass_fractions_per_layer]
    return (
        numpy.asarray(mass_fractions),
        mass_fractions_per_layer,
        mass_fraction_names,
        quantification_parameters,
    )


def _calc_mass_fraction_fp(
    peak_area: numpy.ndarray,
    ref_fit_peak_area: float,
    ref_fit_mass_fraction: float,
    norm_factors: Union[float, numpy.ndarray],
) -> numpy.ndarray:
    """Mass fraction calculation for fundamental parameters."""
    factor = ref_fit_mass_fraction / ref_fit_peak_area
    return factor * (norm_factors * peak_area)


def _calc_mass_fraction_int(
    peak_area: numpy.ndarray,
    ref_fit_peak_area: float,
    ref_fit_mass_fraction: float,
    peak_area_int: float,
    ref_fit_peak_area_int: float,
) -> numpy.ndarray:
    """Mass fraction calculation for internal reference."""
    factor_num = ref_fit_mass_fraction * ref_fit_peak_area_int
    factor_denom = peak_area_int * ref_fit_peak_area
    return (factor_num / factor_denom) * peak_area


def _calc_mass_fractions(
    groups: List[str],
    peak_group_areas: numpy.ndarray,
    wresult: dict,
    fitresult: dict,
    calc_mass_fraction: Callable,
) -> Tuple[List[numpy.ndarray], List[List[numpy.ndarray]], List[str]]:
    """Iterate over groups and calculate mass fractions + per-layer fractions."""
    mass_fraction_names = []
    mass_fractions = []
    nlayers = len(wresult["layerlist"])
    mass_fractions_per_layer = [[] for _ in range(nlayers)]

    for name, peak_area in zip(groups, peak_group_areas):
        if name.lower().startswith("scatter"):
            continue

        mass_fraction_names.append(name)
        ref_fit_peak_area = fitresult["result"][name]["fitarea"]

        # Overall mass fraction
        ref_fit_mass_fraction = wresult["mass fraction"][name]
        mass_fractions.append(
            calc_mass_fraction(
                name, peak_area, ref_fit_peak_area, ref_fit_mass_fraction
            )
        )

        # Mass fraction per layer
        for layer, layer_mass_fractions in zip(
            wresult["layerlist"], mass_fractions_per_layer
        ):
            ref_fit_mass_fraction = wresult[layer]["mass fraction"][name]
            layer_mass_fractions.append(
                calc_mass_fraction(
                    name,
                    peak_area,
                    ref_fit_peak_area,
                    ref_fit_mass_fraction,
                    layer=layer,
                )
            )

    return mass_fractions, mass_fractions_per_layer, mass_fraction_names



from ewokscore import Task
from ewokscore.model import BaseInputModel
from ewokscore.model import BaseOutputModel
from pydantic import Field

from ..nexus_utils import wait_scan_finished


class Inputs(BaseInputModel):
    filename: str = Field(
        description="Bliss dataset HDF5 file name.", examples=["/data/dataset.h5"]
    )
    scan_number: int = Field(description="Scan number.", examples=[1, 2])


class Outputs(BaseOutputModel):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )


class PickScan(Task, input_model=Inputs, output_model=Outputs):
    """Select a single Bliss scan."""

    def run(self):
        filename = self.inputs.filename
        scan_number = self.inputs.scan_number
        bliss_scan_uri = f"{filename}::/{scan_number}.1"
        wait_scan_finished(bliss_scan_uri)
        self.outputs.bliss_scan_uri = bliss_scan_uri



from typing import List

import numpy
from ewokscore import Task
from ewokscore.model import BaseOutputModel
from ewoksdata.data.hdf5.config import guess_dataset_config
from pydantic import Field

from ...io.hdf5 import split_h5uri
from .. import nexus_utils
from ..hdf5_utils import create_hdf5_link
from ..hdf5_utils import link_bliss_scan
from ..models import FluoInputs
from .sum_utils import sum_spectra_from_hdf5


class Inputs(FluoInputs):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )

    detector_names: List[str] = Field(
        description="NXdetector group name.",
        min_length=1,
        examples=[["fx_nano_det0", "fx_nano_det1"]],
    )

    output_root_uri: str = Field(
        description="Output root URI.",
        examples=[
            "/results/dataset.h5",
            "/results/dataset.h5::/1.1",
            "/results/dataset.h5::/1.1/sumspectra",
        ],
    )

    xrf_spectra_uri_template: str = Field(
        default="instrument/{}/data",
        description="HDF5 group URI template of an XRF detector relative to `bliss_scan_uri`.",
    )

    detector_normalization_template: str = Field(
        default="1./<instrument/{}/live_time>",
        description="Expression to evaluate and multiply each spectrum with before summing. "
        "Templates marked with '<...>' are relative to `bliss_scan_uri`.",
    )

    output_detector_name: str = Field(default="mcasum")


class Outputs(BaseOutputModel):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )
    output_root_uri: str = Field(
        description="Output NXentry URI.", examples=["/results/dataset.h5::/1.1"]
    )
    detector_name: str = Field(
        description="Name of the detector containing the summed XRF spectra."
    )
    xrf_spectra_uri_template: str = Field(
        description="HDF5 group URI template of an XRF detector relative to `output_root_uri`.",
    )


class SumXrfSpectra(Task, input_model=Inputs, output_model=Outputs):
    """Add single-scan XRF spectra from multiple detectors."""

    def run(self):
        start_time = nexus_utils.now()

        _, scan_h5path = split_h5uri(self.inputs.bliss_scan_uri)

        sumdetector_name = self.inputs.output_detector_name

        with nexus_utils.save_in_ewoks_process(
            self.inputs.output_root_uri,
            start_time,
            process_config={
                "detector_normalization_template": self.inputs.detector_normalization_template
            },
            default_levels=(scan_h5path, "sumspectra"),
        ) as (process_group, already_existed):
            outentry = process_group.parent
            if not already_existed:
                sum_spectra: numpy.ndarray = sum_spectra_from_hdf5(
                    self.inputs.bliss_scan_uri,
                    self.inputs.xrf_spectra_uri_template,
                    self.inputs.detector_normalization_template,
                    self.inputs.detector_names,
                )

                link_bliss_scan(outentry, self.inputs.bliss_scan_uri, retry_timeout=0)

                nxdata = nexus_utils.create_nxdata(
                    process_group, "mcasum", signal="data"
                )

                dataset_kwargs = guess_dataset_config(
                    scan_shape=(sum_spectra.shape[0],),
                    detector_shape=(sum_spectra.shape[1],),
                    dtype=sum_spectra.dtype,
                )
                dset = nxdata.create_dataset("data", data=sum_spectra, **dataset_kwargs)
                dset.attrs["interpretation"] = "spectrum"

                nxdetector = outentry["instrument"].create_group(sumdetector_name)
                nxdetector.attrs["NX_class"] = "NXdetector"
                create_hdf5_link(nxdetector, "data", dset)
                create_hdf5_link(outentry["measurement"], sumdetector_name, dset)

            output_root_uri = f"{outentry.file.filename}::{outentry.name}"

        self.outputs.bliss_scan_uri = output_root_uri
        self.outputs.detector_name = sumdetector_name
        self.outputs.xrf_spectra_uri_template = self.inputs.xrf_spectra_uri_template
        self.outputs.output_root_uri = output_root_uri



from pathlib import Path

import h5py
import numpy

from ...xrffit.batch._save import XrfResultWriter
from ...xrffit.fit.types import QuantificationParameters
from ...xrffit.fit.types import XRFBatchFitResult
from ...xrffit.pymca_config import PyMcaXrfConfiguration
from ..utils import h5content


def test_xrf_result_writer(tmp_path: Path):
    result = _generate_example_xrf_batch_fit()
    filename = tmp_path / "test.h5"
    configuration = PyMcaXrfConfiguration()

    with XrfResultWriter(f"{filename}::/1.1", default_group="fit") as writer:
        writer.save_configuration(configuration)

        for i in range(2):
            start = i * result.num_spectra
            stop = start + result.num_spectra
            writer.save_fit_result(result, start, stop)

    with h5py.File(filename) as f:
        actual = h5content(f)

    expected = {
        "1.1": {
            "@attrs": {"NX_class", "default"},
            "fit": {
                "@attrs": {"NX_class", "default"},
                "configuration": {
                    "@attrs": {"NX_class"},
                    "data@shape": (),
                    "date@shape": (),
                    "type@shape": (),
                },
                "program@shape": (),
                "quantification": {
                    "@attrs": {"NX_class"},
                    "DetectorArea@attrs": {"units"},
                    "DetectorArea@shape": (),
                    "DetectorDistance@attrs": {"units"},
                    "DetectorDistance@shape": (),
                    "Flux@attrs": {"units"},
                    "Flux@shape": (),
                    "I0@shape": (),
                    "ReferenceElement@shape": (),
                    "ReferenceTransitions@shape": (),
                    "SolidAngle@attrs": {"units"},
                    "SolidAngle@shape": (),
                    "Time@attrs": {"units"},
                    "Time@shape": (),
                },
                "results": {
                    "@attrs": {"NX_class", "default"},
                    "derivatives": {
                        "@attrs": {"NX_class", "axes"},
                        "Cu_Ka@attrs": {"interpretation"},
                        "Cu_Ka@shape": (1024,),
                        "Fe_Ka@attrs": {"interpretation"},
                        "Fe_Ka@shape": (1024,),
                        "Zn_Ka@attrs": {"interpretation"},
                        "Zn_Ka@shape": (1024,),
                        "energy@shape": (1024,),
                        "gain@attrs": {"interpretation"},
                        "gain@shape": (1024,),
                        "zero@attrs": {"interpretation"},
                        "zero@shape": (1024,),
                    },
                    "fit": {
                        "@attrs": {"NX_class", "auxiliary_signals", "axes", "signal"},
                        "energy@shape": (1024,),
                        "data@shape": (4, 1024),
                        "data@attrs": {"interpretation"},
                        "model@shape": (4, 1024),
                        "model@attrs": {"interpretation"},
                        "residuals@shape": (4, 1024),
                        "residuals@attrs": {"interpretation"},
                    },
                    "mass_fractions": {
                        "@attrs": {"NX_class", "auxiliary_signals", "signal"},
                        "Cu@shape": (4,),
                        "Fe@shape": (4,),
                        "Zn@shape": (4,),
                    },
                    "mass_fractions_layer1": {
                        "@attrs": {"NX_class", "auxiliary_signals", "signal"},
                        "Cu@shape": (4,),
                        "Fe@shape": (4,),
                        "Zn@shape": (4,),
                    },
                    "parameters": {
                        "@attrs": {"NX_class", "auxiliary_signals", "signal"},
                        "Cu_Ka@shape": (4,),
                        "Fe_Ka@shape": (4,),
                        "Zn_Ka@shape": (4,),
                        "gain@shape": (4,),
                        "zero@shape": (4,),
                    },
                    "uncertainties": {
                        "@attrs": {"NX_class", "auxiliary_signals", "signal"},
                        "Cu_Ka@shape": (4,),
                        "Fe_Ka@shape": (4,),
                        "Zn_Ka@shape": (4,),
                        "gain@shape": (4,),
                        "zero@shape": (4,),
                    },
                },
                "version@shape": (),
            },
        },
        "@attrs": {"NX_class", "default"},
    }
    assert actual == expected


def _generate_example_xrf_batch_fit() -> XRFBatchFitResult:
    rng = numpy.random.default_rng(seed=42)

    Nspectra = 2
    Nchan = 1024
    Nfit = 800

    raw_channels = numpy.arange(Nchan)
    raw_energies = raw_channels * 0.010
    channels = numpy.linspace(50, 850, Nfit).astype(int)
    energies = raw_energies[channels]

    spectra = rng.poisson(
        lam=50 + 500 * numpy.exp(-0.5 * ((raw_energies - 6.4) / 0.15) ** 2),
        size=(Nspectra, Nchan),
    )
    model = numpy.zeros_like(spectra, dtype=float)
    for i in range(Nspectra):
        model[i] = (
            45
            + 480 * numpy.exp(-0.5 * ((raw_energies - 6.4) / 0.14) ** 2)
            + 200 * numpy.exp(-0.5 * ((raw_energies - 8.0) / 0.20) ** 2)
        )
    residuals = spectra - model

    parameter_names = ["Fe_Ka", "Cu_Ka", "Zn_Ka", "zero", "gain"]
    parameters = numpy.array(
        [[1200.0, 1150.0], [800.0, 820.0], [450.0, 430.0], [40.0, 42.0], [0.010, 0.011]]
    )
    uncertainties = parameters * 0.05

    mass_fraction_names = ["Fe", "Cu", "Zn"]
    mass_fractions = numpy.array([[0.55, 0.53], [0.30, 0.32], [0.15, 0.15]])
    mass_fractions_per_layer = [mass_fractions.copy()]

    Nparameters = len(parameter_names)
    derivatives = rng.normal(scale=1e-3, size=(Nparameters, Nchan))

    quant_params = QuantificationParameters(
        I0=1.0e10,
        Time=10.0,
        Flux=1.0e9,
        DetectorDistance=5.0,
        DetectorArea=0.5,
        SolidAngle=0.02,
        ReferenceElement="Fe",
        ReferenceTransitions="K",
    )

    return XRFBatchFitResult(
        parameters=parameters,
        uncertainties=uncertainties,
        derivatives=derivatives,
        parameter_names=parameter_names,
        channels=channels,
        energies=energies,
        raw_channels=raw_channels,
        raw_energies=raw_energies,
        spectra=spectra,
        model=model,
        residuals=residuals,
        mass_fractions=mass_fractions,
        mass_fractions_per_layer=mass_fractions_per_layer,
        mass_fraction_names=mass_fraction_names,
        quantification_parameters=quant_params,
    )



from ewoksorange.gui.owwidgets.meta import ow_build_opts
from ewoksorange.gui.owwidgets.threaded import OWEwoksWidgetOneThread
from ewoksorange.gui.widgets.parameter_form import ParameterForm

from . import serialize
from .data_viewer import DataViewer


class OWMeshWidget(OWEwoksWidgetOneThread, **ow_build_opts):
    def __init__(self):
        super().__init__()
        self._init_control_area()
        self._init_main_area()

    def _default_inputs_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())
        self._update_input_widgets()
        self._update_output_widgets()

    def handleNewSignals(self) -> None:
        self._update_input_widgets()
        super().handleNewSignals()

    def task_output_changed(self):
        self._update_output_widgets()

    def _init_control_area(
        self, stack: bool = False, multidetector: bool = False, mosaic: bool = False
    ):
        super()._init_control_area()
        self._default_inputs_form = ParameterForm(parent=self.controlArea)
        defaults = {"output_root_uri": "raw_data.h5"}
        values = self.get_default_input_values(include_missing=True, defaults=defaults)
        self.update_default_inputs(**values)

        parameters = {
            "output_filename": {
                "label": "HDF5 filename",
                "value_for_type": "",
                "select": "file",
            },
            "emission_line_groups": {
                "label": "Line groups",
                "value_for_type": "",
                "serialize": serialize.strings_serializer,
                "deserialize": serialize.strings_deserializer,
            },
            "rois": {
                "label": "Regions of interest",
                "value_for_type": "",
                "serialize": serialize.rois_serializer,
                "deserialize": serialize.rois_deserializer,
            },
            "energy": {
                "label": "Energy (keV)",
                "value_for_type": 0.0,
            },
            "shape": {
                "label": "Scan shape",
                "value_for_type": "",
                "serialize": serialize.shape_serializer,
                "deserialize": serialize.shape_deserializer,
            },
            "expo_time": {
                "label": "Exposure time (sec)",
                "value_for_type": 0.0,
            },
            "flux": {
                "label": "Flux (1/sec)",
                "value_for_type": 0.0,
            },
            "counting_noise": {
                "label": "Counting Noise",
                "value_for_type": False,
            },
            "integral_type": {
                "label": "Data as integers",
                "value_for_type": False,
            },
        }

        if mosaic:
            parameters["mosaic"] = {
                "label": "Mosaic chunks",
                "value_for_type": "",
                "serialize": serialize.shape_serializer,
                "deserialize": serialize.shape_deserializer,
            }

        if stack:
            parameters["nscans"] = {
                "label": "Number of scans",
                "value_for_type": 1,
            }

        if multidetector:
            parameters["ndetectors"] = {
                "label": "Number of detectors",
                "value_for_type": 0,
            }

        for name, kw in parameters.items():
            self._default_inputs_form.addParameter(
                name,
                value=values[name],
                value_change_callback=self._default_inputs_changed,
                **kw,
            )

    def _init_main_area(self):
        super()._init_main_area()
        layout = self._get_main_layout()

        self._viewer = DataViewer(parent=self.mainArea)
        self._viewer.setVisible(True)
        layout.addWidget(self._viewer)
        layout.setStretchFactor(self._viewer, 1)

        self._update_output_widgets()

    def _update_input_widgets(self):
        dynamic = self.get_dynamic_input_names()
        for name in self.get_input_names():
            self._default_inputs_form.set_parameter_enabled(name, name not in dynamic)

    def _update_output_widgets(self):
        self._refresh_output_file()

    def _refresh_output_file(self):
        filename = self.get_task_input_value("output_filename")
        if filename:
            self._viewer.updateFile(filename)



from ewoksfluo.gui.mesh_widget import OWMeshWidget
from ewoksfluo.tasks.example_data.tasks import MeshStackMultiDetector


class OWMeshStackMultiDetector(OWMeshWidget, ewokstaskclass=MeshStackMultiDetector):
    name = "Mesh stack (multi detector)"
    description = "XRF test data of a stack of identical scans with multiple detectors"

    def _init_control_area(self):
        super()._init_control_area(stack=True, multidetector=True, mosaic=False)



from contextlib import contextmanager
from multiprocessing import Queue
from typing import Generator
from typing import Optional

from .buffers import AbstractOutputBuffer
from .buffers import ExternalOutputBuffer
from .buffers import PyMcaOutputBuffer
from .handlers import NexusOutputHandler
from .handlers import QueueOutputHandler


@contextmanager
def queue_outputbuffer_context(
    queue: Queue,
    sendid: int,
    destinationid: int,
    nscans: Optional[int] = None,
    scan_index: Optional[int] = None,
    diagnostics: bool = False,
    figuresofmerit: bool = True,
) -> Generator[ExternalOutputBuffer, None, None]:
    with QueueOutputHandler(
        queue, sendid, destinationid, nscans, scan_index
    ) as handler:
        yield ExternalOutputBuffer(
            handler, diagnostics=diagnostics, figuresofmerit=figuresofmerit
        )


@contextmanager
def outputbuffer_context(
    output_root_uri: str,
    diagnostics: bool = False,
    figuresofmerit: bool = True,
    output_handler: str = "nexus",
    open_options: Optional[dict] = None,
) -> Generator[AbstractOutputBuffer, None, None]:
    if open_options is None:
        open_options = {}

    if output_handler == "nexus":
        if diagnostics:
            default_group = "fit"
        else:
            default_group = "parameters"
        with NexusOutputHandler(
            output_root_uri, default_group=default_group, **open_options
        ) as handler:
            yield ExternalOutputBuffer(
                handler, diagnostics=diagnostics, figuresofmerit=figuresofmerit
            )
    elif output_handler == "pymca":
        # Handler and buffer are coupled
        yield PyMcaOutputBuffer(
            output_root_uri,
            diagnostics=diagnostics,
            figuresofmerit=figuresofmerit,
            **open_options,
        )
    else:
        raise ValueError(f"Unknown output handler {output_handler}")



import pathlib
from functools import lru_cache
from typing import Dict
from typing import Generator
from typing import List
from typing import Optional
from typing import Tuple

import h5py
import numpy

from ..io import hdf5
from ..tasks.example_data.xrf_spectra import EmissionLineGroup
from ..tasks.example_data.xrf_spectra import ScatterLineGroup
from ..tasks.example_data.xrf_spectra import xrf_spectra
from ..xrffit.pymca_config import PyMcaXrfConfiguration


def h5content(group: hdf5.GroupType):
    info = dict()
    if group.attrs:
        info["@attrs"] = set(group.attrs)
    for name, item in group.items():
        if hdf5.is_group(item):
            info[name] = h5content(item)
        else:
            if item.attrs:
                info[f"{name}@attrs"] = set(item.attrs)
            info[f"{name}@shape"] = item.shape
    return info


@lru_cache
def generate_data(
    npoints_per_scan: int,
    energy: float,
    tmp_path: Optional[pathlib.Path] = None,
    samefile: bool = True,
    nscans: int = 1,
    ndetectors: int = 1,
    points_with_negative_peaks: Tuple[int, ...] = tuple(),
    quantification: Optional[str] = None,
    num_layers: int = 1,
) -> Tuple[
    List[List[str]], numpy.ndarray, Dict[str, numpy.ndarray], PyMcaXrfConfiguration
]:
    xrf_spectra_uris = list()
    parameters = dict()
    spectra = list()

    fit_background = bool(points_with_negative_peaks)

    for detector in range(ndetectors):
        det_xrf_spectra_uris = list()
        xrf_spectra_uris.append(det_xrf_spectra_uris)
        det_spectra = list()
        spectra.append(det_spectra)

        for scan, filename in enumerate(
            _generate_scan_filenames(tmp_path, samefile, nscans), 1
        ):
            linegroups = [
                EmissionLineGroup(
                    "Si", "K", _generate_counts(300, npoints_per_scan, scan)
                ),
                EmissionLineGroup(
                    "Al", "K", _generate_counts(400, npoints_per_scan, scan)
                ),
                EmissionLineGroup(
                    "Cl",
                    "K",
                    _generate_counts(
                        200,
                        npoints_per_scan,
                        scan,
                        negative_indices=points_with_negative_peaks,
                    ),
                ),
                EmissionLineGroup(
                    "Pb", "M", _generate_counts(500, npoints_per_scan, scan)
                ),
                EmissionLineGroup(
                    "P", "K", _generate_counts(200, npoints_per_scan, scan)
                ),
                EmissionLineGroup(
                    "S", "K", _generate_counts(600, npoints_per_scan, scan)
                ),
                EmissionLineGroup(
                    "Ca", "K", _generate_counts(500, npoints_per_scan, scan)
                ),
                EmissionLineGroup(
                    "Ti", "K", _generate_counts(400, npoints_per_scan, scan)
                ),
                EmissionLineGroup(
                    "Ce", "L", _generate_counts(500, npoints_per_scan, scan)
                ),
                EmissionLineGroup(
                    "Fe", "K", _generate_counts(1000, npoints_per_scan, scan)
                ),
            ]
            scattergroups = [
                ScatterLineGroup(
                    "Peak000", _generate_counts(100, npoints_per_scan, scan)
                ),
                ScatterLineGroup(
                    "Compton000", _generate_counts(100, npoints_per_scan, scan)
                ),
            ]
            if detector == 0:
                for group in linegroups:
                    lst = parameters.setdefault(f"{group.element}_{group.name}", list())
                    lst.extend(group.counts)
                for group in scattergroups:
                    lst = parameters.setdefault(f"{group.prefix}_{group.name}", list())
                    lst.extend(group.counts)

            _spectra, configuration = xrf_spectra(
                linegroups,
                scattergroups,
                energy=energy,
                fit_background=fit_background,
                quantification=quantification,
                num_layers=num_layers,
            )
            det_spectra.extend(_spectra)

            if not filename:
                continue

            with h5py.File(filename, mode="a") as h5file:
                scan_name = f"/{scan}.1"
                spectra_name = f"{scan_name}/measurement/mca{detector:02d}"
                energy_name = f"{scan_name}/instrument/positioners_start/energy"

                h5file[spectra_name] = _spectra
                if energy_name not in h5file:
                    h5file[energy_name] = energy
                    h5file[energy_name].attrs["units"] = "keV"

                det_xrf_spectra_uris.append(f"{filename}::{spectra_name}")

    # For each detector: list of spectra and peak areas for all points in all scans
    spectra = numpy.asarray(spectra)
    parameters = {name: numpy.asarray(values) for name, values in parameters.items()}
    return xrf_spectra_uris, spectra, parameters, configuration


def _generate_scan_filenames(
    tmp_path: Optional[pathlib.Path], samefile: bool, nscans: int
) -> Generator[Optional[str], None, None]:
    if tmp_path is None:
        for scan in range(1, nscans + 1):
            yield None
    elif samefile:
        tmp_path.mkdir(parents=True, exist_ok=True)
        for scan in range(1, nscans + 1):
            yield str(tmp_path / "spectra.h5")
    else:
        tmp_path.mkdir(parents=True, exist_ok=True)
        for scan in range(1, nscans + 1):
            yield str(tmp_path / f"spectra{scan}.h5")


def _generate_counts(
    start_counts: int,
    npoints_per_scan: int,
    scan: int,
    negative_indices: Tuple[int, ...] = tuple(),
) -> List[int]:
    step = 50
    total_step = npoints_per_scan * step
    start = start_counts + (scan - 1) * total_step
    stop = start + total_step
    values = list(range(start, stop, step))
    for i in negative_indices:
        values[i] = -values[i]
    return values



from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

from PyMca5.PyMcaIO import HDF5Stack1D
from PyMca5.PyMcaPhysics.xrf import FastXRFLinearFit
from PyMca5.PyMcaPhysics.xrf import McaAdvancedFitBatch

from ...io.hdf5 import split_h5uri
from ..pymca_config import PyMcaXrfConfiguration
from ..pymca_config import modify_pymca_configuration
from ..pymca_config import pymca_configdict_from_file
from ..pymca_config import pymca_configdict_from_model
from ..pymca_config import pymca_configdict_to_model
from ..pymca_config import temporary_pymca_configdict_to_file


def perform_batch_fit(
    xrf_spectra_uris: Sequence[str],
    pymca_configuration: Union[str, PyMcaXrfConfiguration],
    output_buffer,
    energy: Optional[float] = None,
    energy_multiplier: Optional[float] = None,
    mlines: Optional[dict] = None,
    quantification: Union[Dict[str, float], bool, None] = None,
    fast_fitting: bool = False,
) -> None:
    """Fit XRF spectra in batch with one primary beam energy. Least-square fitting.
    If you intend a linear fit, modify the configuration:

        - Get current energy calibration with "Load From Fit"
        - Enable: Perform a Linear Fit
        - Disable: Stripping
        - Strip iterations = 0

    Fast linear least squares:

          - Use SNIP instead of STRIP

    :param xrf_spectra_uris: spectra to fit.
    :param pymca_configuration: configuration file to use.
    :param output_buffer: object that receives the fit results.
    :param energy: primary beam energy. Defaults to None.
    :param mlines: elements (keys) which M line group must be replaced by some M subgroups (values). Defaults to None.
    :param quantification: save massfractions. Defaults to False.
    :param fast_fitting: use SNIP instead of STRIP. Defaults to False.
    :param diagnostics: fit model and residuals. Defaults to False.
    :param figuresofmerit: chi-square value and other figures of merit. Defaults to False.
    :param energy_multiplier: adds a higher energy bound equal to energy*energy_multiplier to include high-energy peaks. Default: no bound is added.
    :raises ValueError: if all paths in filelist do not have the same HDF5 group as parent.
    """
    if not pymca_configuration:
        raise RuntimeError("PyMca configuration is empty")
    if isinstance(pymca_configuration, str):
        pymca_configuration = pymca_configdict_to_model(
            pymca_configdict_from_file(pymca_configuration)
        )

    modify_pymca_configuration(
        pymca_configuration,
        energy=energy,
        mlines=mlines,
        quantification=quantification,
        fast_fitting=fast_fitting,
        energy_multiplier=energy_multiplier,
    )

    filelist, selection, scanlist = _parse_hdf5_uris(xrf_spectra_uris)
    try:
        if fast_fitting:
            _fast_fit_xrf_spectra_uris(
                filelist,
                selection,
                scanlist,
                pymca_configuration,
                output_buffer,
                quantification=quantification,
            )
        else:
            _fit_xrf_spectra_uris(
                filelist,
                selection,
                scanlist,
                pymca_configuration,
                output_buffer,
                quantification=quantification,
            )
    except Exception as e:
        raise RuntimeError(f"PyMca fitting failed for {xrf_spectra_uris}") from e


def _fast_fit_xrf_spectra_uris(
    filelist: List[str],
    selection: Dict[str, List[str]],
    scanlist: List[str],
    pymca_configuration: PyMcaXrfConfiguration,
    output_buffer,
    quantification: Union[Dict[str, float], bool, None] = None,
):
    batch = FastXRFLinearFit.FastXRFLinearFit()
    stack = HDF5Stack1D.HDF5Stack1D(filelist, selection, scanlist=scanlist)
    kwargs = {
        "y": stack,
        "configuration": pymca_configdict_from_model(pymca_configuration),
        "concentrations": bool(quantification),
        "refit": 1,  # needed to avoid negative peak areas
        "outbuffer": output_buffer,
    }
    with output_buffer.saveContext():
        batch.fitMultipleSpectra(**kwargs)


def _fit_xrf_spectra_uris(
    filelist: List[str],
    selection: Dict[str, List[str]],
    scanlist: List[str],
    pymca_configuration: PyMcaXrfConfiguration,
    output_buffer,
    quantification: Union[Dict[str, float], bool, None] = None,
):
    selection["entry"] = scanlist
    kwargs = {
        "filelist": filelist,
        "selection": selection,
        "concentrations": bool(quantification),
        "fitfiles": 0,
        "fitconcfile": 0,
        "outbuffer": output_buffer,
    }
    with temporary_pymca_configdict_to_file(
        pymca_configdict_from_model(pymca_configuration)
    ) as cfgfilename:
        batch = McaAdvancedFitBatch.McaAdvancedFitBatch(cfgfilename, **kwargs)
        with output_buffer.saveContext():
            batch.processList()


def _parse_hdf5_uris(
    uris: Sequence[str],
) -> Tuple[List[str], Dict[str, List[str]], List[str]]:
    split_results = list(zip(*(split_h5uri(uri) for uri in uris)))
    filelist, datasetlist = split_results
    filelist = list(set(filelist))
    if len(filelist) != 1:
        raise ValueError("cannot handle scans in different files")
    entries, datasets = zip(*(_parse_hdf5_dataset(dataset) for dataset in datasetlist))
    datasets = list(set(datasets))
    if len(datasets) != 1:
        raise ValueError("dataset location is each scan must be the same for all scans")
    selection = {"y": datasets[0]}
    scanlist = list(entries)
    return filelist, selection, scanlist


def _parse_hdf5_dataset(dataset: str) -> Tuple[str, str]:
    parts = [s for s in dataset.split("/") if s]
    return parts[0], "/" + "/".join(parts[1:])



"""Different implementations of the core XRF fit function.

- `ewoksfluo.xrffit.fit.fast_fit.fit_xrf_spectra`: solve a linear system of equations to fit several spectra.
- `ewoksfluo.xrffit.fit.slow_fit.fit_xrf_spectra`: fit several XRF spectra one-by-one.
- `ewoksfluo.xrffit.fit.fast_fit_native.fit_xrf_spectra`: same as `fast_fit` but using the PyMca batch API for fast fitting.
- `ewoksfluo.xrffit.fit.slow_fit_native.fit_xrf_spectra`: same as `slow_fit` but using the PyMca batch API for slow fitting.
- `ewoksfluo.xrffit.fit.fast_fit_native.fit_xrf_spectra_legacy`: same as `fast_fit_native` but before HDF5 support.
- `ewoksfluo.xrffit.fit.slow_fit_native.fit_xrf_spectra_legacy`: same as `slow_fit_native` but before HDF5 support.

All functions can be called through `ewoksfluo.xrffit.fit.fit_xrf_spectra`.
"""

from typing import Optional

import numpy

from ..pymca_config import PyMcaXrfConfiguration
from . import fast_fit
from . import fast_fit_native
from . import slow_fit
from . import slow_fit_native
from .types import XRFBatchFitResult


def fit_xrf_spectra(
    xrf_spectra: numpy.ndarray,
    configuration: PyMcaXrfConfiguration,
    quantification: Optional[bool] = None,
    individual_weights: Optional[bool] = None,
    live_times: Optional[numpy.ndarray] = None,
    positive_peak_areas: Optional[bool] = None,
    diagnostics: Optional[bool] = None,
    fast_fitting: Optional[bool] = None,
    native_fitting: Optional[bool] = None,
    native_legacy_fitting: Optional[bool] = None,
) -> XRFBatchFitResult:
    """
    :param xrf_spectra: shape `(num_spectra, num_channels)`
    :param configuration: PyMca configuration.
    :param quantification: Calculate mass fractions from peak area's.
    :param individual_weights: When fitting with weights, use the weight of each
                                spectrum (slow) instead of the average weight.
    :param live_times: 1D array with one value for each XRF spectrum.
                       Only used for mass fractions with fundamental parameters.
    :param positive_peak_areas:
    :param diagnostics: fit model and residuals.
    :param fast_fitting: Fast fitting means fit all spectra by solving a single linear system of equations.
    :param native_fitting: Use native PyMca batch processing.
    :param native_legacy_fitting: Use legacy native PyMca batch processing.
    """
    if fast_fitting:
        if native_legacy_fitting:
            _fit = fast_fit_native.fit_xrf_spectra_legacy
        elif native_fitting:
            _fit = fast_fit_native.fit_xrf_spectra
        else:
            _fit = fast_fit.fit_xrf_spectra
    else:
        if native_legacy_fitting:
            _fit = slow_fit_native.fit_xrf_spectra_legacy
        elif native_fitting:
            _fit = slow_fit_native.fit_xrf_spectra
        else:
            _fit = slow_fit.fit_xrf_spectra

    return _fit(
        xrf_spectra,
        configuration,
        quantification=quantification,
        individual_weights=individual_weights,
        live_times=live_times,
        positive_peak_areas=positive_peak_areas,
        diagnostics=diagnostics,
    )



import logging
from typing import List

from ewokscore import Task
from ewokscore.model import BaseOutputModel
from pydantic import Field

from ..math import format_expression_template
from ..models import FluoInputs
from .sum_utils import detector_weight_iterator
from .sum_utils import save_summed_xrf_results

_logger = logging.getLogger(__name__)


class Inputs(FluoInputs):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )

    detector_names: List[str] = Field(
        description="NXdetector group name.",
        min_length=1,
        examples=[["fx_nano_det0", "fx_nano_det1"]],
    )

    output_root_uri: str = Field(
        description="Output root URI.",
        examples=[
            "/results/dataset.h5",
            "/results/dataset.h5::/1.1",
            "/results/dataset.h5::/1.1/sum",
        ],
    )

    xrf_results_uris: List[str] = Field(
        description="Output NXprocess results URIs.",
        examples=[
            [
                "/results/dataset.h5::/1.1/fit/mca1/results",
                "/results/dataset.h5::/1.1/fit/mca2/results",
            ]
        ],
    )

    detector_normalization_template: str = Field(
        default="1./<instrument/{}/live_time>",
        description="Expression to evaluate and multiply each spectrum with before summing. "
        "Templates marked with '<...>' are relative to `bliss_scan_uri`.",
    )


class Outputs(BaseOutputModel):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )
    output_root_uri: str = Field(
        description="Output NXentry URI.", examples=["/results/dataset.h5::/1.1"]
    )
    xrf_results_uri: str = Field(
        description="Output NXprocess results URI.",
        examples=["/results/dataset.h5::/1.1/sum/results"],
    )


class SumXrfResults(Task, input_model=Inputs, output_model=Outputs):
    """Add single-scan XRF results of multiple detectors."""

    def run(self) -> None:
        weight_expressions = [
            format_expression_template(
                self.inputs.detector_normalization_template, name
            )
            for name in self.inputs.detector_names
        ]
        for weight_expression, detector_name in zip(
            weight_expressions, self.inputs.detector_names
        ):
            _logger.info(
                "Detector %r weights for summing XRF results: %s",
                detector_name,
                weight_expression,
            )
        detector_weights = detector_weight_iterator(
            self.inputs.bliss_scan_uri, weight_expressions
        )

        process_config = {
            "detector_normalization_template": self.inputs.detector_normalization_template
        }
        self.outputs.xrf_results_uri = save_summed_xrf_results(
            self.inputs.xrf_results_uris,
            detector_weights,
            self.inputs.output_root_uri,
            process_config,
        )
        self.outputs.bliss_scan_uri = self.inputs.bliss_scan_uri
        self.outputs.output_root_uri = self.inputs.output_root_uri



import numpy
import pytest

from ..math.regular_grid import ScatterDataInterpolator


@pytest.mark.parametrize("ndatadim", [0, 1, 2])
def test_regular_regridding(ndatadim):
    data_shape = tuple(range(10, 10 + ndatadim))

    # Plane: z = ax + by + c
    a, b, c = 3, 2, 1

    state = numpy.random.RandomState(42)
    N_points = 100
    x = state.uniform(-10, 10, N_points)
    y = state.uniform(-10, 10, N_points)
    z = a * x + b * y + c
    z = _add_data_dimensions(z, data_shape)

    interpolator = ScatterDataInterpolator(
        [x, y], ["x", "y"], ["um", "um"], method="linear"
    )
    interpolated_z = interpolator.regrid(z)

    expected_z = (
        a * interpolator.expanded_grid_coordinates[:, 0]
        + b * interpolator.expanded_grid_coordinates[:, 1]
        + c
    )
    expected_z = _add_data_dimensions(expected_z, data_shape)
    expected_z = expected_z.reshape(interpolated_z.shape)

    assert numpy.nanmax(numpy.abs(interpolated_z - expected_z)) < 1e-9


def _add_data_dimensions(array: numpy.ndarray, data_shape: tuple) -> numpy.ndarray:
    reshaped_array = array[(...,) + (numpy.newaxis,) * len(data_shape)]
    return numpy.tile(reshaped_array, data_shape)



from ..math import scan_shape


def test_scan_shape():
    nfast = 49
    nslow = 53

    npoints = [nfast * nslow, (nfast + 1) * nslow]
    equations = ["nfast * nslow", "(nfast + 1) * nslow"]

    solution = scan_shape.solve_scan_shape(npoints, equations)
    assert solution == {"nfast": nfast, "nslow": nslow}


def test_scan_shape_over_determined():
    nfast = 49
    nslow = 53

    npoints = [nfast * nslow, (nfast + 1) * nslow, 2 * (nfast + 1) * nslow]
    equations = ["nfast * nslow", "(nfast + 1) * nslow", "2 * (nfast + 1) * nslow"]

    solution = scan_shape.solve_scan_shape(npoints, equations)
    assert solution == {"nfast": nfast, "nslow": nslow}


def test_scan_shape_under_determined():
    nfast = 49
    nslow = 53

    npoints = [nfast * nslow]
    equations = ["nfast * nslow"]

    solution = scan_shape.solve_scan_shape(npoints, equations)
    assert solution is None


def test_solve_nnls_posint():
    nfast = 49
    nslow = 53

    equations_dict = {
        "npoints1": "nfast * nslow",
        "npoints2": "(nfast + 1) * nslow",
        "npoints3": "2 * (nfast + 1) * nslow",
        "npoints4": "npoints1",
    }

    values_dict = {
        "npoints1": nfast * nslow,
        "npoints2": (nfast + 1) * nslow,
        "npoints3": 2 * (nfast + 1) * nslow,
        "npoints4": nfast * nslow,
    }

    solution = scan_shape.solve_nnls_posint(equations_dict, values_dict)
    assert solution == {"nfast": nfast, "nslow": nslow}



import logging
from contextlib import contextmanager

import numpy

from .. import resource_utils


def test_job_parameters(system_4cpus_8gb):
    # 8 GB array
    shape = (1024, 1024, 1024)
    dtype = numpy.float64

    njobs = 1
    narrays_in_parallel, chunk_size_dim0 = resource_utils.job_parameters(
        njobs, shape, dtype
    )
    assert narrays_in_parallel == 1
    assert chunk_size_dim0 == 1024

    # Check MEM limit with chunking
    njobs = 2
    narrays_in_parallel, chunk_size_dim0 = resource_utils.job_parameters(
        njobs, shape, dtype
    )
    assert narrays_in_parallel == 2
    assert chunk_size_dim0 == 512

    # Check MEM limit without chunking
    njobs = 2
    narrays_in_parallel, chunk_size_dim0 = resource_utils.job_parameters(
        njobs, shape, dtype, chunking_possible=False
    )
    assert narrays_in_parallel == 1
    assert chunk_size_dim0 == 1024

    # Check CPU limit
    njobs = 16
    narrays_in_parallel, chunk_size_dim0 = resource_utils.job_parameters(
        njobs, shape, dtype
    )
    assert narrays_in_parallel == 3
    assert chunk_size_dim0 == 341


def test_log_required_memory(system_4cpus_8gb, caplog):
    # 4 GB array
    shape = (1024, 1024, 1024)
    dtype = numpy.float32
    level = logging.INFO
    expected = "Memory required to process data: 4.00 GB (available: 8.00 GB)"

    with _assert_logging(caplog, level, expected):
        resource_utils.log_required_memory("process data", shape, dtype)

    # 16 GB array
    shape = (2048, 1024, 1024)
    dtype = numpy.float64
    level = logging.WARNING
    expected = "Memory required to process data: 16.00 GB (available: 8.00 GB)"

    with _assert_logging(caplog, level, expected):
        resource_utils.log_required_memory("process data", shape, dtype)


@contextmanager
def _assert_logging(caplog, level, message):
    with caplog.at_level(level):
        yield

    messages = {r.message for r in caplog.records}
    assert message in messages



import os
import stat

import h5py
import pytest

from ..io import hdf5

skip_if_root = pytest.mark.skipif(
    os.geteuid() == 0, reason="Test requires non-root permissions"
)


def test_join_h5url():
    assert hdf5.join_h5url("filename.h5", "") == "filename.h5::/"
    assert hdf5.join_h5url("filename.h5::/", "/") == "filename.h5::/"

    assert hdf5.join_h5url("filename.h5", "1.1") == "filename.h5::/1.1"
    assert hdf5.join_h5url("filename.h5::/", "/1.1/") == "filename.h5::/1.1"


@pytest.fixture
def hdf5_files(tmp_path):
    root_path = tmp_path / "root.h5"
    ext_path = tmp_path / "ext.h5"

    with h5py.File(ext_path, "w") as ext:
        ext["/group/data"] = [4, 5, 6]

    with h5py.File(root_path, "w") as root:
        root["/internal/data"] = [1, 2, 3]
        root["/external1"] = h5py.ExternalLink("ext.h5", "/")
        root["/external2/data"] = h5py.ExternalLink("ext.h5", "/group/data")

    os.chmod(ext_path, stat.S_IREAD)
    yield root_path, ext_path
    os.chmod(ext_path, stat.S_IWRITE | stat.S_IREAD)


def test_read_access(hdf5_files):
    root_path, _ = hdf5_files

    with hdf5.FileModExtAccess(root_path, mode="a") as root:
        root.set_external_access(mode="r")

        assert "/internal/data" in root
        dset = root["/internal/data"]
        assert dset[()].tolist() == [1, 2, 3]

        assert "/external1/group/data" in root
        dset = root["/external1/group/data"]
        assert dset[()].tolist() == [4, 5, 6]

        assert "/external2/data" in root
        dset = root["/external2/data"]
        assert dset[()].tolist() == [4, 5, 6]

        assert "/internal/notexisting" not in root
        assert "/external1/group" in root
        assert "/external1/group/notexisting" not in root
        assert "/external2/notexisting" not in root


def test_write_access(hdf5_files):
    root_path, _ = hdf5_files

    with hdf5.FileModExtAccess(root_path, mode="a") as root:
        root.set_external_access(mode="r")
        dset = root["/internal/data"]
        dset[0] = 99
        assert dset[()].tolist() == [99, 2, 3]

        dset = root["/external1/group/data"]
        with pytest.raises(OSError):
            dset[0] = 99
        assert dset[()].tolist() == [4, 5, 6]

        dset = root["/external2/data"]
        with pytest.raises(OSError):
            dset[0] = 99
        assert dset[()].tolist() == [4, 5, 6]


@skip_if_root
def test_create_access(hdf5_files):
    root_path, ext_path = hdf5_files

    with hdf5.FileModExtAccess(root_path, mode="a") as root:
        root["/internal/new"] = 10
        assert root["/internal/new"][()] == 10

        with pytest.raises(PermissionError):
            root["/external1/group/new"] = 20

    os.chmod(ext_path, stat.S_IWRITE | stat.S_IREAD)

    with hdf5.FileModExtAccess(root_path, mode="a") as root:
        root["/external1/group/new"] = 30
        assert root["/external1/group/new"][()] == 30


@skip_if_root
def test_delete_access(hdf5_files):
    root_path, ext_path = hdf5_files

    with hdf5.FileModExtAccess(root_path, mode="a") as root:
        del root["/internal/data"]
        assert "/internal/data" not in root

        with pytest.raises(PermissionError):
            del root["/external1/group/data"]

        del root["/external2/data"]
        assert "/external2/data" not in root

    os.chmod(ext_path, stat.S_IWRITE | stat.S_IREAD)

    with hdf5.FileModExtAccess(root_path, mode="a") as root:
        del root["/external1/group/data"]

        assert "/external1/group/data" not in root



from typing import List
from typing import Union

from ewokscore import Task
from ewokscore.model import BaseInputModel
from ewokscore.model import BaseOutputModel
from pydantic import Field

from ...io.convert import spec_to_bliss


class Inputs(BaseInputModel):
    input_filename: str = Field(
        description="SPEC master file name.", examples=["/data/specfile.dat"]
    )
    output_filename: str = Field(
        description="Bliss dataset HDF5 file name.",
        examples=["/results/dataset.h5"],
    )
    scan_numbers: Union[int, List[int], None] = Field(
        default=None, description="Scan numbers.", examples=[1, [1, 2, 3]]
    )
    subscan_numbers: Union[int, List[int], None] = Field(
        default=None, description="Subscan numbers.", examples=[1, [1, 2]]
    )


class Outputs(BaseOutputModel):
    output_filename: str = Field(
        description="Bliss dataset HDF5 file name.",
        examples=["/results/dataset.h5"],
    )


class SpecToBliss(Task, input_model=Inputs, output_model=Outputs):
    """Convert SPEC master file to BLISS dataset HDF5 file."""

    def run(self):
        spec_filename = self.inputs.input_filename
        bliss_filename = self.inputs.output_filename
        spec_to_bliss(
            spec_filename,
            bliss_filename,
            scans=self.inputs.scan_numbers,
            subscans=self.inputs.subscan_numbers,
            mode="a",
        )
        self.outputs.output_filename = bliss_filename



import json

from ewoksorange.gui.owwidgets.meta import ow_build_opts
from ewoksorange.gui.owwidgets.threaded import OWEwoksWidgetOneThread
from ewoksorange.gui.widgets.parameter_form import ParameterForm

from ..io.hdf5 import split_h5uri
from .data_viewer import DataViewer


class OWFitWidget(OWEwoksWidgetOneThread, **ow_build_opts):
    def __init__(self):
        super().__init__()
        self._multidetector = False
        self._init_control_area()
        self._init_main_area()

    def _init_control_area(self, stack: bool = False, multidetector: bool = False):
        self._multidetector = multidetector

        super()._init_control_area()
        self._default_inputs_form = ParameterForm(parent=self.controlArea)
        defaults = {"output_root_uri": "processed_data.h5"}
        values = self.get_default_input_values(include_missing=True, defaults=defaults)
        self.update_default_inputs(**values)

        parameters = dict()

        if stack:
            parameters["bliss_scan_uris"] = {
                "label": "List of scan URIs",
                "value_for_type": "",
                "serialize": json.dumps,
                "deserialize": json.loads,
            }
        else:
            parameters["bliss_scan_uri"] = {
                "label": "Scan URI",
                "value_for_type": "",
                "select": "h5group",
            }
        parameters["output_root_uri"] = {
            "label": "Output URI",
            "value_for_type": "",
            "select": "h5group",
        }
        parameters["process_uri_template"] = {
            "label": "Process URI template",
            "value_for_type": "",
        }

        if multidetector:
            parameters["detector_names"] = {
                "label": "List of detector names",
                "value_for_type": "",
                "serialize": json.dumps,
                "deserialize": json.loads,
            }
        else:
            parameters["detector_name"] = {
                "label": "Detector name",
                "value_for_type": "",
            }
        parameters["xrf_spectra_uri_template"] = {
            "label": "XRF spectra URI template",
            "value_for_type": "",
        }

        if multidetector:
            parameters["configs"] = {
                "label": "List of PyMCA config URIs",
                "value_for_type": "",
                "serialize": json.dumps,
                "deserialize": json.loads,
            }
        else:
            parameters["config"] = {
                "label": "PyMCA config URI",
                "value_for_type": "",
                "select": "file",
            }

        parameters["energy_name"] = {
            "label": "Beam energy positioner name",
            "value_for_type": "",
        }
        parameters["energy_uri_template"] = {
            "label": "Beam energy positioner template",
            "value_for_type": "",
            "select": "h5group",
        }

        parameters["quantification"] = {
            "label": "Quantification",
            "value_for_type": False,
        }
        parameters["energy_multiplier"] = {
            "label": "Energy multiplier",
            "value_for_type": 0.0,
        }
        parameters["fast_fitting"] = {
            "label": "Fast",
            "value_for_type": False,
        }
        parameters["diagnostics"] = {
            "label": "Diagnostics",
            "value_for_type": False,
        }

        for name, kw in parameters.items():
            self._default_inputs_form.addParameter(
                name,
                value=values[name],
                value_change_callback=self._default_inputs_changed,
                **kw,
            )

    def _default_inputs_changed(self):
        self.update_default_inputs(**self._default_inputs_form.get_parameter_values())
        self._update_input_widgets()
        self._update_output_widgets()

    def handleNewSignals(self) -> None:
        self._update_input_widgets()
        super().handleNewSignals()

    def task_output_changed(self):
        self._update_output_widgets()

    def _init_main_area(self):
        super()._init_main_area()
        layout = self._get_main_layout()

        self._viewer = DataViewer(parent=self.mainArea)
        self._viewer.setVisible(True)
        layout.addWidget(self._viewer)
        layout.setStretchFactor(self._viewer, 1)

        self._update_output_widgets()

    def _update_input_widgets(self):
        dynamic = self.get_dynamic_input_names()
        for name in self.get_input_names():
            self._default_inputs_form.set_parameter_enabled(name, name not in dynamic)

    def _update_output_widgets(self):
        self._refresh_output_file()

    def _get_output_filenames(self):
        if self._multidetector:
            uris = self.get_task_output_value("xrf_results_uris")
            if not uris:
                return list()
            return [split_h5uri(uri)[0] for uri in uris]
        else:
            uri = self.get_task_output_value("xrf_results_uri")
            if not uri:
                return list()
            return [split_h5uri(uri)[0]]

    def _refresh_output_file(self):
        filenames = self._get_output_filenames()
        for filename in filenames:
            self._viewer.updateFile(filename)



from ewoksfluo.gui.fit_widget import OWFitWidget
from ewoksfluo.tasks.fit.tasks import FitStackMultiDetector


class OWFitStackMultiDetector(OWFitWidget, ewokstaskclass=FitStackMultiDetector):
    name = "Fit stack (multi detector)"
    description = "Fit a stack of identical scans with multiple detectors"

    def _init_control_area(self):
        super()._init_control_area(stack=True, multidetector=True)



from typing import List

import h5py
import numpy

from ..io.blissconcat import concatenate_bliss_scans


def test_concatenate_basic(tmp_path):
    infile = tmp_path / "dataset.h5"
    outfile = tmp_path / "concat.h5"

    values1 = numpy.arange(5)
    values2 = numpy.arange(5, 12)
    _create_test_scan(infile, "1.1", values1, scalars=["a", "b", "c"])
    _create_test_scan(infile, "2.1", values2, scalars=["a", "b"])

    result_uri = concatenate_bliss_scans(
        [f"{infile}::/1.1", f"{infile}::/2.1"],
        f"{outfile}::/1.1",
        virtual_axes={"sy": "<samy> + <sampy>"},
        axes_units={"sampy": "um", "sy": "cm"},
    )
    assert result_uri == f"{outfile}::/1.1"

    with h5py.File(outfile, "r") as f:
        samy = f["/1.1/measurement/samy"][()]
        sampy = f["/1.1/measurement/sampy"][()]
        sy = f["/1.1/measurement/sy"][()]
        positioners = set(f["/1.1/instrument/positioners"])

    expected = numpy.arange(12)  # mm
    assert numpy.allclose(samy, expected)

    expected = numpy.arange(12)  # um
    assert numpy.allclose(sampy, expected)

    expected = numpy.arange(12) / 10 + numpy.arange(12) / 10000  # cm
    assert numpy.allclose(sy, expected)

    assert positioners == {"samy", "sampy", "sy", "a", "b"}


def _create_test_scan(
    infile: str, scan: str, values: numpy.ndarray, scalars: List[str]
) -> None:
    with h5py.File(infile, "a") as h5file:
        grp = h5file.create_group(scan)
        grp.attrs["NX_class"] = "NXentry"

        instrument = grp.create_group("instrument")
        instrument.attrs["NX_class"] = "NXinstrument"

        pos = instrument.create_group("positioners")
        pos.attrs["NX_class"] = "NXcollection"

        meas = grp.create_group("measurement")
        meas.attrs["NX_class"] = "NXcollection"

        dset = pos.create_dataset("samy", data=values)
        dset.attrs["units"] = "mm"
        meas["samy"] = h5py.SoftLink(dset.name)

        dset = pos.create_dataset("sampy", data=values)
        # Missing unit
        # dset.attrs["units"] = "um"
        meas["sampy"] = h5py.SoftLink(dset.name)

        for scalar in scalars:
            pos[scalar] = 0



import re
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional

from scipy.optimize import least_squares


def solve_scan_shape(
    num_scan_points: List[int], equations: List[str]
) -> Optional[Dict[str, int]]:
    """Compute the scan shape based on a number of scan points and their parametrization.

    :params num_scan_points: A list of scan points.
    :param equations: A parametrization of the list of scan points.
    :returns: If the solver, a dictionary containing the solved integer values for each variable.
    """
    if len(equations) != len(num_scan_points):
        raise ValueError("Number of equations and number scan points must be the same")
    keys = [f"npoints{i}" for i in range(len(equations))]
    equations_dict = dict(zip(keys, equations))
    values_dict = dict(zip(keys, num_scan_points))

    def estimate_initial_values(nunknowns) -> List[int]:
        min_value = min(values_dict.values())
        return [max(1.0, min_value ** (1 / nunknowns))] * nunknowns

    return solve_nnls_posint(equations_dict, values_dict, estimate_initial_values)


def solve_nnls_posint(
    equations_dict: Dict[str, str],
    values_dict: Dict[str, int],
    estimate_initial_values: Optional[Callable[[int], List[int]]] = None,
) -> Optional[Dict[str, int]]:
    """Solve a system of non-linear equations with strictly positive integer constraints, supporting over-determined systems.

    :param equations_dict: A dictionary where keys are equation names and values are the equations as strings.
    :param values_dict: A dictionary where keys match those in equations_dict and values are the known results.
    :returns: If the solver, a dictionary containing the solved integer values for each variable.
    :raises ValueError: If the keys in equations_dict and values_dict do not match.
    """
    if set(equations_dict.keys()) != set(values_dict.keys()):
        raise ValueError("Mismatch between equations and values dictionary keys")
    if not equations_dict:
        return

    unknowns = set()
    for eq in equations_dict.values():
        unknowns.update(re.findall(r"[a-zA-Z_]\w*", eq))
    unknowns -= set(values_dict.keys())  # Remove constants
    unknowns = sorted(unknowns)

    def residuals(vars: List[float]) -> List[float]:
        local_vars = dict(zip(unknowns, vars))
        residuals = []
        for key, equation in equations_dict.items():
            # Evaluate each equation with current variable values
            residual = (
                eval(equation, {}, {**local_vars, **values_dict}) - values_dict[key]
            )
            residuals.append(residual)
        return residuals

    nunknowns = len(unknowns)
    if estimate_initial_values:
        initial_guess = estimate_initial_values(nunknowns)
    else:
        initial_guess = [1.0] * nunknowns

    bounds = (1, [float("inf")] * nunknowns)
    result = least_squares(residuals, initial_guess, bounds=bounds)
    final_values = list(map(round, result.x))

    final_residuals = residuals(final_values)
    if any(final_residuals):
        return

    return dict(zip(unknowns, final_values))



import logging
import multiprocessing
from concurrent.futures import ProcessPoolExecutor
from contextlib import ExitStack
from queue import Queue
from typing import Dict
from typing import List
from typing import Optional
from typing import Sequence
from typing import Set
from typing import Tuple
from typing import Union

from ... import h5_subprocess
from ...io.hdf5 import join_h5url
from ...resource_utils import job_parameters
from ...xrffit.legacy import outputbuffer_context
from ...xrffit.legacy import perform_batch_fit
from ...xrffit.legacy import queue_outputbuffer_context
from ...xrffit.legacy.handlers import NexusOutputHandler
from ...xrffit.legacy.handlers import consume_handler_queue
from ...xrffit.legacy.handlers import stop_queue
from ...xrffit.pymca_config import PyMcaXrfConfiguration
from ..hdf5_utils import get_dataset_shape_and_dtype

logger = logging.getLogger(__name__)


def fit_single(
    bliss_scan_uri: str,
    xrf_spectra_uri_template: str,
    output_root_uri: str,
    process_uri_template: str,
    detector_name: str,
    pymca_configuration: Union[str, PyMcaXrfConfiguration],
    energy: Optional[float] = None,
    quantification: Union[Dict[str, float], bool, None] = None,
    energy_multiplier: Optional[float] = None,
    fast_fitting: bool = False,
    diagnostics: bool = False,
    figuresofmerit: Optional[bool] = None,
    **_,
) -> Tuple[str, str]:
    """Fitting of one scan with one detector.

    Returns the URL of the fit result and the URL to the NXentry group.
    """
    if figuresofmerit is None:
        figuresofmerit = diagnostics
    output_uri_template = join_h5url(output_root_uri, process_uri_template)
    output_root_uri = output_uri_template.format(detector_name)
    with outputbuffer_context(
        output_root_uri,
        diagnostics=diagnostics,
        figuresofmerit=figuresofmerit,
    ) as output_buffer:
        if output_buffer.already_existed:
            logger.warning("%s already exists", output_buffer.xrf_results_uri)
        else:
            xrf_spectra_uri = (
                f"{bliss_scan_uri}/{xrf_spectra_uri_template.format(detector_name)}"
            )
            perform_batch_fit(
                xrf_spectra_uris=[xrf_spectra_uri],
                pymca_configuration=pymca_configuration,
                output_buffer=output_buffer,
                energy=energy,
                energy_multiplier=energy_multiplier,
                quantification=quantification,
                fast_fitting=fast_fitting,
            )
        return output_buffer.xrf_results_uri, output_buffer.output_root_uri


def fit_multi(
    bliss_scan_uris: Sequence[str],
    xrf_spectra_uri_template: str,
    output_root_uri: str,
    process_uri_template: str,
    detector_names: Sequence[str],
    pymca_configurations: Sequence[Union[str, PyMcaXrfConfiguration]],
    energies: Optional[Sequence[Optional[float]]] = None,
    quantification: Union[Dict[str, float], bool, None] = None,
    energy_multiplier: Optional[float] = None,
    fast_fitting: bool = False,
    diagnostics: bool = False,
    figuresofmerit: Optional[bool] = None,
    **_,
) -> Tuple[List[str], str]:
    """Parallelized fitting of a stack of identical scan with multiple detectors.

    Returns the URL's of the fit result (one URL per detector) and the URL to the NXentry group.
    """
    nscans = len(bliss_scan_uris)
    if energies:
        if len(energies) != nscans:
            raise ValueError(f"Requires {nscans} energies, one for each scan")
    else:
        energies = [None] * nscans
    ndetectors = len(detector_names)
    if len(pymca_configurations) != ndetectors:
        raise ValueError(
            f"Requires {ndetectors} pymca configurations, one for each detector"
        )

    if figuresofmerit is None:
        figuresofmerit = diagnostics
    if diagnostics:
        default_group = "fit"
    else:
        default_group = "parameters"

    xrf_results_uris: List[str] = []
    next_output_root_uri: Optional[str] = None

    with ExitStack() as stack:
        mp_context = multiprocessing.get_context(method="spawn")
        ctx = mp_context.Manager()
        manager = stack.enter_context(ctx)
        queue = manager.Queue()

        arguments: List[Tuple[tuple, dict, dict]] = list()
        output_handlers: Dict[int, NexusOutputHandler] = dict()
        queue_sendids: Set[int] = set()

        output_uri_template = join_h5url(output_root_uri, process_uri_template)
        output_uris = [
            output_uri_template.format(detector_name)
            for detector_name in detector_names
        ]
        if len(set(output_uris)) != len(output_uris):
            raise ValueError(
                "Add a place-holder '{}' for the detector name in the output URI template"
            )
        xrf_spectra_uris = [
            [
                f"{bliss_scan_uri}/{xrf_spectra_uri_template.format(detector_name)}"
                for bliss_scan_uri in bliss_scan_uris
            ]
            for detector_name in detector_names
        ]
        if len(set(xrf_spectra_uris[0])) != len(xrf_spectra_uris[0]):
            raise ValueError(
                "Add a place-holder '{}' for the detector name in the XRF spectra URI template"
            )

        for destinationid, (
            output_root_uri,
            detector_spectra_uris,
            pymca_configuration,
        ) in enumerate(zip(output_uris, xrf_spectra_uris, pymca_configurations)):
            ctx = NexusOutputHandler(output_root_uri, default_group=default_group)
            output_handler = stack.enter_context(ctx)
            xrf_results_uris.append(output_handler.xrf_results_uri)
            next_output_root_uri = output_handler.output_root_uri
            if output_handler.already_existed:
                logger.warning("%s already exists", output_handler.xrf_results_uri)
                continue
            output_handlers[destinationid] = output_handler

            for i_scan, (xrf_spectra_uri, energy) in enumerate(
                zip(detector_spectra_uris, energies)
            ):
                queue_sendid = destinationid * nscans + i_scan
                queue_sendids.add(queue_sendid)

                buffer_args = (
                    queue,
                    queue_sendid,
                    destinationid,
                    nscans,
                    i_scan,
                )
                buffer_kwargs = {
                    "diagnostics": diagnostics,
                    "figuresofmerit": figuresofmerit,
                }
                fit_kwargs = {
                    "xrf_spectra_uris": [xrf_spectra_uri],
                    "pymca_configuration": pymca_configuration,
                    "energy": energy,
                    "energy_multiplier": energy_multiplier,
                    "quantification": quantification,
                    "fast_fitting": fast_fitting,
                }
                arguments.append(
                    (
                        _fit_main.__module__,
                        _fit_main.__name__,
                        buffer_args,
                        buffer_kwargs,
                        fit_kwargs,
                    )
                )

        if not arguments:
            return xrf_results_uris, next_output_root_uri

        # Sub-processes will send the fit results to the queue
        shape, dtype = get_dataset_shape_and_dtype(xrf_spectra_uri)
        multiplier = 3
        max_workers, _ = job_parameters(
            len(arguments), shape, dtype, multiplier=multiplier, chunking_possible=False
        )

        ctx = ProcessPoolExecutor(
            mp_context=mp_context,
            initializer=h5_subprocess.initializer,
            max_workers=max_workers,
        )
        executor = stack.enter_context(ctx)

        futures = {
            i: executor.submit(h5_subprocess.main, *args)
            for i, args in enumerate(arguments)
        }

        def raise_on_error():
            for i, future in list(futures.items()):
                if future.done():
                    future.result()  # re-raise exception (if any)
                    futures.pop(i)

        # Main process will receive results from the queue and save them in HDF5
        consume_handler_queue(output_handlers, queue, queue_sendids, raise_on_error)

        # Re-raise exceptions (if any)
        for future in futures.values():
            future.result()

    return xrf_results_uris, next_output_root_uri


def _fit_main(
    buffer_args: Tuple[Queue, int, int, Optional[int], int],
    buffer_kwargs: dict,
    fit_kwargs: dict,
) -> None:
    queue, sendid, destinationid, nscans, scan_index = buffer_args
    if nscans == 1:
        nscans = None
        scan_index = None
    try:
        with queue_outputbuffer_context(
            queue,
            sendid,
            destinationid,
            nscans=nscans,
            scan_index=scan_index,
            **buffer_kwargs,
        ) as output_buffer:
            perform_batch_fit(output_buffer=output_buffer, **fit_kwargs)
    finally:
        stop_queue(queue, sendid)



from typing import Dict
from typing import Optional

import numpy
from numpy.typing import DTypeLike


def apply_extendable_deadtime(
    photons,
    elapsed_time,
    max_dt_percentage,
    integral_type: Optional[DTypeLike] = numpy.uint32,
):
    """Number of counts measured by a digital signal processor with extendable a.k.a. paralyzable
    deadtime for a given number of incoming photons."""
    r = photons / elapsed_time
    max_r = r.max()
    tau = -numpy.log(1 - max_dt_percentage / 100) / max_r
    dt = 1 - numpy.exp(-tau * photons / elapsed_time)
    measured = photons * (1 - dt)
    if integral_type is None:
        return measured
    return (measured + 0.5).astype(integral_type)


def apply_dualchannel_signal_processing(
    spectra: numpy.ndarray,
    max_dt_fast=0.1,
    max_dt_slow=20,
    elapsed_time=0.1,
    counting_noise: bool = True,
    integral_type: Optional[DTypeLike] = numpy.uint32,
) -> Dict[str, numpy.ndarray]:
    """MCA spectra with statistics measured by a dual-channel digital signal processor:
    one channel with low deadtime and low energy precision and
    one channel with high deadtime and high energy precision.

    The last dimension of `spectra` is the MCA channel/energy dimension.
    An MCA spectrum (multi-channel analyzer) contains measured photons binned
    by their energy.
    """
    shape = spectra.shape[:-1]
    if integral_type:
        spectra = spectra.astype(integral_type)
    if counting_noise:
        spectra = numpy.random.poisson(spectra)
    photons = spectra.sum(axis=-1)
    elapsed_time = numpy.full(shape, elapsed_time)
    events = apply_extendable_deadtime(
        photons, elapsed_time, max_dt_slow, integral_type
    )
    triggers = apply_extendable_deadtime(photons, elapsed_time, max_dt_fast)
    fractional_dead_time = 1 - events / photons
    event_count_rate = events / elapsed_time
    trigger_live_time = elapsed_time * triggers / photons
    live_time = elapsed_time * events / photons
    trigger_count_rate = triggers / trigger_live_time
    spectra = spectra * (events / photons)[..., numpy.newaxis]
    if integral_type:
        spectra = spectra.astype(integral_type)
    return {
        "spectrum": spectra,
        "events": events,
        "triggers": triggers,
        "event_count_rate": event_count_rate,
        "trigger_count_rate": trigger_count_rate,
        "live_time": live_time,
        "trigger_live_time": trigger_live_time,
        "elapsed_time": elapsed_time,
        "fractional_dead_time": fractional_dead_time,
    }



import h5py
import numpy

from ..tasks.normalization import norm_utils


def test_normalization_coefficient(tmp_path):
    live_time = numpy.random.uniform(low=0.08, high=0.1, size=11)
    iodet = numpy.random.uniform(low=0.08, high=0.1, size=10)
    with h5py.File(str(tmp_path / "dataset.h5"), mode="w") as f:
        f["/1.1/instrument/fx_sxm_det0/live_time"] = live_time
        f["/1.1/instrument/iodet/data"] = iodet
        weight_uri_root = f"{f.filename}::/1.1"

    normalization_expression = "0.1/<instrument/fx_sxm_det0/live_time>*np.nanmean(<instrument/iodet/data>)/<instrument/iodet/data>"

    expected = numpy.empty((11,), dtype=float)
    expected[:10] = 0.1 / live_time[:10] * numpy.mean(iodet[:10]) / iodet[:10]
    expected[-1] = numpy.nan
    coefficient = norm_utils.normalization_coefficient(
        weight_uri_root, normalization_expression
    )
    numpy.testing.assert_array_equal(expected, coefficient)



import logging
from typing import Optional

from ewokscore import Task
from ewokscore.model import BaseOutputModel
from pydantic import Field

from ..models import FluoInputs
from .norm_utils import normalization_coefficient
from .norm_utils import normalization_template
from .norm_utils import save_normalized_xrf_results

_logger = logging.getLogger(__name__)


class Inputs(FluoInputs):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )

    output_root_uri: str = Field(
        description="Output root URI.",
        examples=[
            "/results/dataset.h5",
            "/results/dataset.h5::/1.1",
            "/results/dataset.h5::/1.1/norm",
        ],
    )

    xrf_results_uri: Optional[str] = Field(
        description="Previous XRF NXprocess URI to normalize.",
        examples=["/results/dataset.h5::/1.1/sum/results"],
    )

    normalization_expression: Optional[str] = Field(
        default=None, description="Arithmetic expression to be used for normalization."
    )
    counter_normalization_template: str = Field(
        default="np.nanmean(<instrument/{}/data>)/<instrument/{}/data>",
        description="Arithmetic expression to be used for normalization.",
    )
    counter_name: Optional[str] = Field(
        default=None,
        description="To be used in :math:`counter_normalization_template`.",
    )
    detector_normalization_template: str = Field(
        default="1./<instrument/{}/live_time>",
        description="Arithmetic expression to be used for normalization.",
    )
    detector_name: Optional[str] = Field(
        default=None,
        description="To be used in :math:`detector_normalization_template`.",
    )


class Outputs(BaseOutputModel):
    bliss_scan_uri: str = Field(
        description="Bliss scan URI.", examples=["/data/dataset.h5::/1.1"]
    )
    output_root_uri: str = Field(
        description="Output NXentry URI.", examples=["/results/dataset.h5::/1.1"]
    )
    xrf_results_uri: str = Field(
        description="Output NXprocess results URI.",
        examples=["/results/dataset.h5::/1.1/norm/results"],
    )


class NormalizeXrfResults(
    Task,
    input_model=Inputs,
    output_model=Outputs,
):
    """Normalize XRF results with raw scan counters.

    Typical normalizers are beam monitors and measurement (live) time.
    """

    def run(self):
        expression = self.inputs.normalization_expression
        detector_name = self.inputs.detector_name
        counter_name = self.inputs.counter_name
        if not (expression or detector_name or counter_name):
            _logger.warning(
                "Neither 'normalization_expression', 'counter_name', nor 'detector_name' was specified. Normalization is skipped."
            )
            self.outputs.xrf_results_uri = self.inputs.xrf_results_uri
            self.outputs.bliss_scan_uri = self.inputs.bliss_scan_uri
            self.outputs.output_root_uri = self.inputs.output_root_uri
            return

        counter_normalization_template = self.inputs.counter_normalization_template
        detector_normalization_template = self.inputs.detector_normalization_template

        normalization_expression = normalization_template(
            expression,
            counter_normalization_template,
            counter_name,
            detector_normalization_template,
            detector_name,
        )
        _logger.info("Multiply XRF results with: %s", normalization_expression)
        coefficient = normalization_coefficient(
            self.inputs.bliss_scan_uri, normalization_expression
        )

        process_config = {"normalization_expression": normalization_expression}
        self.outputs.xrf_results_uri = save_normalized_xrf_results(
            self.inputs.xrf_results_uri,
            coefficient,
            self.inputs.output_root_uri,
            process_config,
        )
        self.outputs.bliss_scan_uri = self.inputs.bliss_scan_uri
        self.outputs.output_root_uri = self.inputs.output_root_uri



from typing import List
from typing import Tuple

try:
    from typing import Self
except ImportError:
    from typing_extensions import Self

from ewokscore import Task
from ewokscore.model import BaseOutputModel
from pydantic import Field
from pydantic import model_validator

from ..models import FluoInputs
from ..nexus_utils import wait_scan_finished


class Inputs(FluoInputs):
    filenames: List[str] = Field(
        description="Bliss dataset HDF5 file names.",
        examples=[["/data/dataset1.h5", "/data/dataset2.h5"]],
    )

    scan_ranges: List[Tuple[int, int]] = Field(
        description="Ranges of scan numbers.",
        examples=[[(1, 4), (100, 101)]],
    )

    exclude_scans: List[List[int]] = Field(
        default_factory=list,
        description="Scan numbers to exclude for each range.",
        examples=[[[1, 3], []]],
    )

    @model_validator(mode="after")
    def check_and_expand_lengths(self) -> Self:
        n_files = len(self.filenames)
        n_ranges = len(self.scan_ranges)
        n_exclude = len(self.exclude_scans)

        if n_files != n_ranges:
            raise ValueError("`filenames` and `scan_ranges` must have the same length")

        if n_exclude == 0:
            self.exclude_scans = [[] for _ in range(n_files)]
            return self

        if n_exclude != n_files:
            raise ValueError(
                "`exclude_scans` must have the same length as `filenames` or be empty"
            )

        return self


class Outputs(BaseOutputModel):
    bliss_scan_uris: List[str] = Field(
        description="Bliss scan URI.",
        examples=[
            [
                "/data/dataset1.h5::/2.1",
                "/data/dataset1.h5::/4.1",
                "/data/dataset2.h5::/100.1",
                "/data/dataset2.h5::/101.1",
            ]
        ],
    )


class PickScans(Task, input_model=Inputs, output_model=Outputs):
    """Select a multiple Bliss scan from multiple files."""

    def run(self):
        filenames = self.inputs.filenames
        scan_ranges = self.inputs.scan_ranges
        exclude_scans = self.inputs.exclude_scans

        bliss_scan_uris = []
        for filename, scan_range, excluded_scans in zip(
            filenames, scan_ranges, exclude_scans
        ):
            scan_min, scan_max = scan_range
            excluded_scans = excluded_scans if excluded_scans else []
            for scan_number in range(scan_min, scan_max + 1):
                if scan_number in excluded_scans:
                    continue

                bliss_scan_uri = f"{filename}::/{scan_number}.1"
                wait_scan_finished(bliss_scan_uri)
                bliss_scan_uris.append(bliss_scan_uri)

        self.outputs.bliss_scan_uris = bliss_scan_uris



import os
import pathlib
import warnings
from collections import abc
from contextlib import contextmanager
from typing import Dict
from typing import Generator
from typing import Optional
from typing import Tuple
from typing import Union

import h5py
from silx.io import h5py_utils
from silx.io.url import DataUrl


def split_h5uri(url: Union[str, pathlib.Path, DataUrl]) -> Tuple[str, str]:
    if isinstance(url, DataUrl):
        obj = url
    else:
        obj = DataUrl(str(url))
    return obj.file_path(), obj.data_path() or ""


def join_h5url(root_url: Union[str, pathlib.Path, DataUrl], sub_url: str) -> str:
    file_path, data_path = split_h5uri(root_url)

    while data_path.endswith("/"):
        data_path = data_path[:-1]
    while data_path.endswith("::"):
        data_path = data_path[:-2]

    while sub_url.startswith("/"):
        sub_url = sub_url[1:]
    while sub_url.endswith("/"):
        sub_url = sub_url[:-1]

    return f"{file_path}::{data_path}/{sub_url}"


def is_file(item) -> bool:
    return isinstance(item, (h5py.File, FileModExtAccess))


def is_group(item) -> bool:
    return isinstance(item, (h5py.Group, GroupModExtAccess))


def is_dataset(item) -> bool:
    return isinstance(item, h5py.Dataset)


FileType = Union[h5py.File, "FileModExtAccess"]
GroupType = Union[h5py.Group, "GroupModExtAccess"]
DatasetType = h5py.Dataset


class GroupModExtAccess(abc.MutableMapping):
    """
    Wrapper around h5py.Group that delegates path resolution
    to the owning FileModExtAccess.
    """

    def __init__(self, native_group: h5py.Group, file_wrapper: "FileModExtAccess"):
        self._native_group = native_group
        self._file_wrapper = file_wrapper

    def __getitem__(self, name: str) -> Union[h5py.Dataset, "GroupModExtAccess"]:
        return self._file_wrapper._resolve_from(self._native_group, name)

    def __setitem__(self, name: str, value) -> None:
        parent_name, _, set_name = name.rpartition("/")
        parent = self._file_wrapper._resolve_from(self._native_group, parent_name)
        parent._native_group[set_name] = value

    def __delitem__(self, name: str) -> None:
        parent_name, _, del_name = name.rpartition("/")
        parent = self._file_wrapper._resolve_from(self._native_group, parent_name)
        del parent._native_group[del_name]

    def __iter__(self):
        return iter(self._native_group)

    def __len__(self) -> int:
        return len(self._native_group)

    @property
    def name(self) -> str:
        return self._native_group.name

    @property
    def file(self) -> "FileModExtAccess":
        return self._file_wrapper

    def get(self, name: str, default=None, getclass=False, getlink=False):
        if getclass or getlink:
            return self._native_group.get(
                name, default=default, getclass=getclass, getlink=getlink
            )
        if name in self._native_group:
            return self[name]
        return default

    @property
    def attrs(self):
        return self._native_group.attrs

    def __repr__(self):
        return f'<{type(self).__name__} "{self.name}">'


class FileModExtAccess(GroupModExtAccess):
    """
    Wrapper around an ``h5py_utils.File`` that ensures any ``h5py.ExternalLink``
    is opened with alternative access parameters.
    """

    def __init__(self, filename: Union[str, pathlib.Path], **open_options):
        if not isinstance(filename, (str, pathlib.Path)):
            raise TypeError("filename must be a string")

        self._external_open_options = dict(open_options)
        self._native_file = h5py_utils.File(filename, **open_options)
        self._external_files: Dict[str, FileModExtAccess] = {}
        super().__init__(self._native_file, self)

    def set_external_access(self, **opts):
        self._external_open_options.update(opts)

    @property
    def filename(self) -> str:
        return self._native_file.filename

    def close(self):
        # Close external files then root file
        for root in list(self._external_files.values()):
            try:
                root.close()
            except Exception:
                pass
        self._external_files.clear()
        if self._native_file is not None:
            try:
                self._native_file.close()
            finally:
                self._native_file = None

    def __enter__(self) -> "FileModExtAccess":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        self.close()

    def _open_external(
        self, filename: str, owner_file: h5py_utils.File
    ) -> h5py_utils.File:
        """Open (or return cached) external file. Always opened with read + no locking."""
        if not os.path.isabs(filename):
            base_dir = os.path.dirname(owner_file.filename)
            filename = os.path.join(base_dir, filename)

        key = os.path.realpath(filename)
        if key in self._external_files:
            return self._external_files[key]

        ext_h5 = FileModExtAccess(filename, **self._external_open_options)
        self._external_files[key] = ext_h5
        return ext_h5

    def _resolve_from(
        self, native_parent: Union[h5py_utils.File, h5py.Group], name: str
    ) -> Union[h5py.Dataset, "GroupModExtAccess"]:
        # If absolute, start from the file that owns the parent
        is_absolute = name.startswith("/")
        if is_absolute:
            native_current = native_parent.file
        else:
            native_current = native_parent
        current_file = self

        # Traverse tree structure explicitly with explicit link resolution
        parts = [p for p in name.split("/") if p]
        i = 0
        while i < len(parts):
            part = parts[i]
            try:
                link = native_current.get(part, getlink=True)
            except Exception:
                link = None

            if isinstance(link, h5py.ExternalLink):
                ext_file = self._open_external(link.filename, native_current.file)
                link_parts = [p for p in link.path.split("/") if p]
                # jump into external target path + remaining tail
                parts = link_parts + parts[i + 1 :]
                native_current = ext_file._native_file
                current_file = ext_file
                i = 0
                continue

            if isinstance(link, h5py.SoftLink):
                link_parts = [p for p in link.path.split("/") if p]
                parts = link_parts + parts[i + 1 :]
                link_is_absolute = link.path.startswith("/")
                if link_is_absolute:
                    native_current = native_current.file
                i = 0
                continue

            # Hard link, dataset, group or virtual dataset
            obj = native_current[part]

            i += 1
            if i < len(parts):
                # Need to descend further: obj must be a group
                if isinstance(obj, h5py.Group):
                    native_current = obj
                    continue
                else:
                    raise KeyError(
                        f"Cannot descend into non-group '{part}' while resolving path '{name}'"
                    )
            else:
                # Last component: return dataset or wrapped group
                native_current = obj

        # Return dataset or wrapped group
        if isinstance(native_current, h5py.Group):
            return GroupModExtAccess(native_current, current_file)
        return native_current


class FileReadAccess(FileModExtAccess):
    """Use in cases where you want to read something from an HDF5 which
    might be already open for writing.
    """

    def __init__(self, filename: Union[str, pathlib.Path], mode: str = "r", **kwargs):
        if not isinstance(filename, (str, pathlib.Path)):
            raise TypeError("filename must be a string")
        assert mode == "r", "must be opened read-only"
        try:
            super().__init__(filename, mode=mode, **kwargs)
        except Exception:
            super().__init__(filename, mode="a", **kwargs)
            self.set_external_access(mode="r")


class ReadHdf5File(FileReadAccess):

    def __init__(self, *args, **kwargs):
        warnings.warn(
            "ReadHdf5File is deprecated, please use FileReadAccess instead.",
            category=DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)


@contextmanager
def retry_external_link(
    group: h5py.Group, name: str, item: Optional[Union[h5py.Group, h5py.Dataset]] = None
) -> Generator[Union[h5py.Group, h5py.Dataset], None, None]:
    """The file we save results in is opened in append mode. If we access external links to
    the raw Bliss data we might not have permission to open it in append mode (e.g. restored data).
    """
    warnings.warn(
        "retry_external_link is deprecated, please use ReadHdf5File instead to open the file.",
        category=DeprecationWarning,
        stacklevel=2,
    )

    if item is None:
        try:
            item = group[name]
        except (KeyError, ValueError):
            item = None
    if item is None:
        link = group.get(name, getlink=True)
        if not isinstance(link, h5py.ExternalLink):
            raise RuntimeError(f"Broken link '{group.name}/{name}'")

        external_filename = link.filename
        if not os.path.isabs(external_filename):
            parent_dir = os.path.dirname(group.file.filename)
            external_filename = os.path.abspath(
                os.path.normpath(os.path.join(parent_dir, external_filename))
            )

        with h5py_utils.File(external_filename, mode="r") as f:
            yield f[link.path]
    else:
        yield item


def is_same_file(
    url1: Union[str, pathlib.Path, DataUrl], url2: Union[str, pathlib.Path, DataUrl]
) -> bool:
    filename1, _ = split_h5uri(url1)
    filename2, _ = split_h5uri(url2)
    filename1 = os.path.abspath(os.path.normpath(filename1))
    filename2 = os.path.abspath(os.path.normpath(filename2))
    return filename1 == filename2



from queue import Empty
from queue import Queue
from typing import Callable
from typing import Dict
from typing import Optional
from typing import Set
from typing import Tuple

import numpy

from ..queue_messages import DataMsg
from ..queue_messages import DatasetMsg
from ..queue_messages import GroupMsg
from ..queue_messages import QueueMsg
from ..queue_messages import StopMsg
from ..queue_messages import is_DataMsg
from ..queue_messages import is_DatasetMsg
from ..queue_messages import is_GroupMsg
from ..queue_messages import is_StopMsg
from .abstract import AbstractDataHandler
from .abstract import AbstractOutputHandler
from .nexus import NexusDataHandlerType
from .nexus import NexusOutputHandler


class _QueueDataHandler(AbstractDataHandler):
    def __init__(
        self,
        group: str,
        name: str,
        queue: Queue,
        sendid: int,
        destinationid: int,
        scan_index: Optional[int],
    ) -> None:
        self._queue = queue
        self._sendid = sendid
        self._destinationid = destinationid
        self._group = group
        self._name = name
        self._scan_index = scan_index

    def add_points(self, value: numpy.ndarray) -> None:
        msg = DataMsg(
            self._sendid,
            self._destinationid,
            self._group,
            self._name,
            value,
            self._scan_index,
        )
        self._queue.put(msg)


class QueueOutputHandler(AbstractOutputHandler):
    def __init__(
        self,
        queue: Queue,
        sendid: int,
        destinationid: int,
        nscans: Optional[int],
        scan_index: Optional[int],
    ):
        self._queue = queue
        self._sendid = sendid
        self._destinationid = destinationid
        self._nscans = nscans
        self._scan_index = scan_index

    def __enter__(self) -> "QueueOutputHandler":
        return self

    def __exit__(self, *args):
        stop_queue(self._queue, self._sendid)

    def create_group(self, name: str, data: dict) -> None:
        msg = GroupMsg(self._sendid, self._destinationid, name, data)
        self._queue.put(msg)

    def create_nxdata_handler(
        self,
        group_name: str,
        name: str,
        npoints: int,
        attrs: Optional[dict] = None,
    ) -> _QueueDataHandler:
        if attrs is None:
            attrs = {}
        msg = DatasetMsg(
            self._sendid,
            self._destinationid,
            group_name,
            name,
            npoints,
            attrs,
            self._nscans,
        )
        self._queue.put(msg)
        return _QueueDataHandler(
            group_name,
            name,
            self._queue,
            self._sendid,
            self._destinationid,
            self._scan_index,
        )


def consume_handler_queue(
    output_handlers: Dict[int, NexusOutputHandler],
    queue: Queue,
    all_sendids: Set[int],
    raise_on_error: Callable[[], None],
):
    datasets: Dict[Tuple[int, str, str], NexusDataHandlerType] = dict()
    stopped_sendids: Set[int] = set()
    while True:
        raise_on_error()
        try:
            msg: QueueMsg = queue.get(timeout=0.5)
        except Empty:
            continue

        if is_StopMsg(msg):
            stopped_sendids.add(msg.sendid)
            if all_sendids == stopped_sendids:
                return
            continue

        if is_GroupMsg(msg):
            destid = msg.destinationid
            output_handler = output_handlers[destid]
            output_handler.create_group(msg.group, msg.data)
            continue

        if is_DatasetMsg(msg):
            group, name, destid = msg.group, msg.name, msg.destinationid
            output_handler = output_handlers[destid]
            dataset_key = (destid, group, name)
            if dataset_key not in datasets:
                datasets[dataset_key] = output_handler.create_nxdata_handler(
                    group, name, msg.npoints, msg.attrs, msg.nscans
                )
            continue

        if is_DataMsg(msg):
            group, name, destid = msg.group, msg.name, msg.destinationid
            output_handler = output_handlers[destid]
            dataset_key = (destid, group, name)
            datasets[dataset_key].add_points(msg.value, msg.scan_index)
            continue

        raise ValueError(f"Unknown command {msg.cmd}")


def stop_queue(queue: Queue, sendid: int):
    msg = StopMsg(sendid)
    queue.put(msg)
